
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "xproto.h"
#include "xprascii.h"

#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};


char _prog_name[] = "<QDOS XPR-ASCII v2.30>";
static QVS(_qcfx_ , "QCFX>>01\x00\x08XPRASCII\x00\x042.30\xFF\xFF");

char *find_option(char *, char);

static struct xpr_option op[] =
{
    {"ASCII options:", XPRO_HEADER, NULL, 0},
    {"EOL convert (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"Pad blank lines (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"Inter-char delay:", XPRO_LONG, NULL, 0},
    {"Inter-line delay:", XPRO_LONG, NULL, 0},
};

static char option_e[4], option_p[4], option_c[8], option_l[8];

CheckSetup (IO)
PXIO IO;
{
    if (IO->xpr_data == NULL)
    {
	if ((IO->xpr_data = (PXPD) malloc (sizeof (XPD))) == NULL)
	{
	    ioerr (IO, "Not enough memory");
	    return ERROR;
	}
	setmem (IO->xpr_data, sizeof (XPD), '\0');
	XD->EOL = 0;
	XD->Pad = 0;
	XD->CDelay = 0;
	XD->LDelay = 0;
    }
    return OK;
}

/**********************************************************
 **********************************************************/

long XProtocolSetup (IO)
PXIO IO;
{

    char buf[80], *p;
    int i, len;

    if (CheckSetup (IO) == ERROR)
    {
	return XPRS_FAILURE;
    }

    /* If options string passed by comm prog, use it; else prompt user */
    if (IO->xpr_filename)
    {
	strcpy (buf, IO->xpr_filename);
    }
    else
    {
	/* If xpr_options() implemented by comm program, use it */
	if (IO->xpr_extension >= 1 && IO->xpr_options)
	{
	    op[1].xpro_value = option_e;
	    op[1].xpro_length = sizeof (option_e);
	    *option_e = (XD->EOL) ? 'Y' : 'N';
	    *(option_e + 1) = '\0';
	    op[2].xpro_value = option_p;
	    op[2].xpro_length = sizeof (option_p);
	    *option_p = (XD->Pad) ? 'Y' : 'N';
	    *(option_p + 1) = '\0';
	    op[3].xpro_value = option_c;
	    op[3].xpro_length = sizeof (option_c);
	    sprintf(option_c, "%d", XD->CDelay);
	    op[4].xpro_value = option_l;
	    op[4].xpro_length = sizeof (option_l);
	    sprintf(option_l, "%d", XD->LDelay);

	    /* Convert Y/N used elsewhere into "yes"/"no" required by spec */
	    for (i = 1; i < 3; ++i)
	    {
		strcpy (op[i].xpro_value, (op[i].xpro_value[0] == 'Y') ? "yes" : "no");
	    }

	    (*IO->xpr_options) (op, 5);

	    /* Convert "yes"/"no" or "on"/"off" into Y/N */
             for (i = 1; i < 3; ++i)
             {
                  if(op[i].xpro_value[0] == 'Y' ||
                    op[i].xpro_value[0] == 'y' ||
                    stricmp (op[i].xpro_value, "yes") == 0 ||
                    stricmp (op[i].xpro_value, "on") == 0)
                    {
                        op[i].xpro_value[0] = 'Y';
                    }
                    else
                    {
                        op[i].xpro_value[0] = 'N';
                    }

             }
            sprintf (buf, "E%c,P%c,C%s,L%s",
	    		*option_e, *option_p, option_c, option_l);

	}
	else
	{
	    /* Start buffer with current settings so user can see/edit them. */
            sprintf (buf, "E%c,P%c,C%d,L%d", 
	    	(XD->EOL) ? 'Y' : 'N',
	    	(XD->Pad) ? 'Y' : 'N',		
		XD->CDelay, XD->LDelay);
	    if (IO->xpr_gets)
		(*IO->xpr_gets) ("XModem options:", buf);
	}
    }
    /* Upshift options string for easier parsing */
    strupr (buf);

    if (p = find_option (buf, 'E'))
    {
	if (*p == 'Y' || *p == 'N')
	    XD->EOL = (*p == 'Y');
	else
	    ioerr (IO, "Invalid E flag ignored; should be Y, N");
    }

    if (p = find_option (buf, 'P'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->Pad = (*p == 'Y');
	}
	else
	    ioerr (IO, "Invalid P flag ignored; should be Y or N");
    }

    if (p = find_option (buf, 'C'))
    {
	i = atoi(p);
	if(i >= 0 && i < 500)
	{
	    XD->CDelay = i;
	}
	else
	    ioerr (IO, "Invalid C flag ignored; (0 - 500)");
    }

    if (p = find_option (buf, 'L'))
    {
	i = atoi(p);
	if(i >= 0 && i < 500)
	{
	    XD->LDelay = i;
	}
	else
	    ioerr (IO, "Invalid L flag ignored; (0 - 500)");
    }
    return XPRS_SUCCESS;
}


ASetup (IO)
PXIO IO;
{
    if (!IO->xpr_update)
    {
	return ERROR;
    }
    if (!IO->xpr_fopen || !IO->xpr_fclose || !IO->xpr_fread
	|| !IO->xpr_fwrite || !IO->xpr_fseek || !IO->xpr_sread
	|| !IO->xpr_swrite)
    {
	ioerr (IO, "Comm prog missing required function(s); see docs");
	return ERROR;
    }

    if (CheckSetup (IO) == ERROR)
    {
	ioerr (IO, "Setup Fails");
	return ERROR;
    }
    XD->Want = 0;
    XD->Last = 0;

    return OK;

}
char *find_option(buf, option)
char *buf;
char option;
{
    while (*buf)
    {
        buf += strspn(buf," ,\t\r\n");
        if (*buf == option)
        {
            return ++buf;
        }
        buf += strcspn(buf," ,\t\r\n");
    }
    return NULL;
}

ioerr (IO, text)
PXIO IO;
char *text;
{
    XPRU u;
    if(IO->xpr_update)
    {
        setmem(&u, sizeof(XPRU), '\0');
        u.xpru_msg = text;
	u.xpru_updatemask = XPRU_MSG;
	(*IO->xpr_update)(&u);
    }
}
