/*
 * Demonstration program from XPR Programming
 *
 * (c) Jonathan Hudson
 */

#define MAIN_MODULE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <qdos.h>

#include <pd/thing.h>
#include <pd/xproto.h>

#include "xprdemo.h"
#include "xprmsg.h"

/*
 * Need a big window
 */

struct WINDOWDEF _condetails =
{
    2, 1, 0, 7, 512, 256, 0, 0
};

void (*_consetup) () = consetup_title;
char _prog_name[] = "XPR Demo II v0.01 Nov 93";
char *_endmsg = NULL;

extern PXPRTHG pXB;

/*
 * Solicits 'Prefs' values from user. A fairly trivial example
 */

XPROptions (struct xpr_option *xpo, int nopt)
{
    int i;
    char txt[64], *p;
    struct xpr_option *o;
    int ival = 1;
    int done = 0;

    sd_clear (pChId1, -1);
    for (o = xpo, i = 0; i < nopt && !done; i++, o++)
    {
	cputs (o->xpro_description, pChId1);
	if (o->xpro_type != XPRO_HEADER)
	{
	    if (o->xpro_type == XPRO_COMMAND)
	    {
		cputs (" [Y/N] ", pChId1);
		io_fline (pChId1, -1, txt, (short) sizeof (txt));
		if (*txt == 'Y' || *txt == 'y')
		{
		    done = 1;

                    /*
                     * COMMANDs require a masking defining the action
                     */
                     
		    ival = (1 << i);
		}
	    }
	    else
	    {
		cputs (" [", pChId1);
		cputs (o->xpro_value, pChId1);
		cputs ("] ", pChId1);
		io_fline (pChId1, -1, txt, (short) sizeof (txt));
		if (*txt)
		{
		    if ((p = strchr (txt, '\n')))
		    {
			*p = '\0';
		    }
		    if (*txt)
		    {
			strcpy (o->xpro_value, txt);
		    }
		}
	    }
	}
	else
	{
	    cputc ('\n', pChId1);
	}
    }
    return ival;
}

/*
 * Variable arguments display message
 */
OutError (char *ctl,)
{
    va_list va;
    char t[80];

    va_start (va, ctl);
    vsprintf (t, ctl, va);
    va_end (va);

    io_sstrg (pChId2, -1, t, (short) strlen (t));
}
cgets(char *p, unsigned int n, int chan)
{
    char *s;
    io_fline (pChId2, -1, p, (short) n);
    if((s = strchr(p, '\n')))
    {
        *s = '\0';
    }
}
    
/*
 * Select a new library, display a list first
 */

SelectXPR ()
{
    int n;
    char *buf;

    buf = malloc ((size_t) 1024);
    if (buf)
    {
	n = XPRFind (buf, 1024);
	if (n > 0)
	{
	    char *p, *q, *s;
            /*
             * Parse name from filename
             */
	    for (p = buf; *p;)
	    {
		q = strstr (p, "xpr");
		p += strlen (p) + 1;
		if (q && *(q + 3))
		{
		    q = q + 3;
		    if ((s = strchr (q, '.')))
		    {
			*s = '\0';
		    }
		    strupr (q);
		    cputs ("    ", pChId2);
		    cputs (q, pChId2);
		    io_sbyte (pChId2, -1, '\n');
		}
	    }

	    cputs ("Library : ", pChId2);
            cgets(MyProtocol, sizeof(MyProtocol), pChId2);
	    if (*MyProtocol)
	    {
		XPRStatus = XPRUse (MyProtocol);
		if (XPRPid == 0)
		{
		    char c;

		    cputs ("Any key to exit", pChId2);
		    io_fbyte (pChId2, -1, &c);
		}
	    }
	}
	else
	{
	    OutError ("Didn't find any XPRs\n");
	}
	free (buf);
    }
}

/*
 * We need to have defined this structure (used by QTPI/QeM) to manage
 * the serial device. We only need to define the first field, the serial
 * channel Id.
 */
 
static SERBLK Sb = {0};

main (int ac, char **av)
{
    int j, iosts;
    short done;
    char FileName[128];
    char action[4];
    int baud = 19200;

    /*
      * XPRD -x Protocol -s serdev -l libdir -b baud
     */

    while ((j = getopt (ac, av, "x:s:b:l:")) != EOF)
    {
	switch (j)
	{
	    case 'x':
		if (optarg)
		{
		    strcpy (MyProtocol, optarg);
		}
		break;
	    case 's':
		if (optarg)
		{
		    strcpy (SerDev, optarg);
		}
		break;
	    case 'l':
		if (optarg)
		{
		    strcpy (LibDir, optarg);
		}
	    case 'b':
		if (optarg)
		{
		    baud = atoi (optarg);
		}
		break;
	}
    }

    pS = &Sb;

    if ((pS->pSerId = io_open (SerDev, 0)) < 0)
    {
	poserr ("Open serial device");
	exit (_oserr);
    }

    mt_baud (baud);

    /*
     * Now, we must find the "XPR Stuff" THING
     */

    GetXPRThing ();

    if (pXB == NULL)
    {
	poserr ("Find XPR Stuff Thing");
	exit (_oserr);
    }

    /*
     * The consoles we use
     */
     
    pChId1 = io_open ("con_442x122a32x20", 0);
    pChId2 = io_open ("con_442x100a32x142", 0);
    sd_bordr (pChId1, -1, WHITE_M4, 1);
    sd_clear (pChId1, -1);
    sd_clear (pChId2, -1);

    /*
     * Now try and load the XPR library
     */

    XPRStatus = XPRUse (MyProtocol);

    for (done = 0; !done && XPRPid;)
    {
	/*
	 * Request an action (and maybe a file name). For receive or monitor,
	 * the file name can be blank if the protocol supports batch modes
	 * (YMODEM, ZMODEM, KERMIT, SEALINK). For a send, the file name
	 * can contain a list, wildcards etc.
	 *
	 * e.g.
	 * Filename: ram2_*_c flp2_*_h flp1_std*_h @ram1_files_list
	 */

	cputs ("Prefs/Send/Receive/Monitor/Library/Quit : ", pChId2);
	cgets (action, sizeof (action), pChId2);

	switch (*action | ' ')
	{
	    case 'l':
		SelectXPR ();
		break;
	    case 'q':
		done = 1;
		break;
	    case 's':
		/*
		 * Send file(s)
		 */
		cputs ("File Name : ", pChId2);
                cgets (FileName, sizeof(FileName), pChId2);                
		if (*FileName)
		{
		    iosts = XPRIO (FileName, 's');
		}
		break;
	    case 'r':
		/*
		 * Receive file(s)
		 */
		cputs ("File Name : ", pChId2);
                cgets (FileName, sizeof(FileName), pChId2);                
		iosts = XPRIO (FileName, 'r');
		break;
	    case 'm':
		/*
		 * Monitor for file(s) (or key) *** ZMODEM ONLY ***
		 * Check if it supports XPRS_HOSTMON
		 */

		if ((XPRStatus & XPRS_HOSTMON) == 0)
		{
		    OutError ("Sorry, can't do that\n");
		}
		else
		{
		    int res;

		    for (res = 0; !res;)
		    {
			char c;

			/*
			 * Scan for character from serial device or console
			 * any console character terminates the HOSTMON loop
			 * If we get a serial character, we use XPRMon to
			 * see if its part of or terminates a ZMODEM sign on.
			 * If it does terminate an XPR sign on, then XPRMon
			 * will start the down load and return 1, so we also
			 * exit this section.
			 */

			if (io_fbyte (pS->pSerId, 5, &c) == 0)
			{
			    io_sbyte (pChId2, -1, c);
			    res = XPRMon (c);
			    if (res)
			    {
				AutoXPR (res);
			    }
			}
			if (io_fbyte (pChId2, 0, &c) == 0)
			{
			    res = 1;
			}
		    }
		}
		break;
	    case 'p':
		DoXPROptions ();
		break;
	    default:
		break;
	}
    }
    /*
     * To tidy up, we free and remove the XPR Stuff THING and kill
     * off the XPR process
     */

    FreeXPRThing ();
    XPRKill ();
    return 0;
}
