/*
 * XPRStuff - The hard work starts here
 */
    
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <qlib.h>
#include <pd/xproto.h>

#include "xprdemo.h"
#include "xprmsg.h"

#define PIPEOK(p1)  (p1) > 0

/*
 * This structure is used by the undocumented spawn function
 */

typedef struct
{
    char *FileName;
    char *ComStr;
    char *InFile;
    int inflag;
    char *OuFile;
    int ouid;
    int JobId;
    int serr;
    unsigned char Prio;
    unsigned char Wait;
}
SJOB, *PSJOB;

static int IPipe = -1;
static int OPipe = -1;

static char *rbuf = "*\030B0x";
static char *abuf;

extern PXPRTHG pXB;

/*
 * Finds the XPRs, into buffer, returns  the number
 * Buffer format is that of getfnl
 * i.e. name1\0name2\0....nameN\0\0
 */
 
XPRFind (char *buf, int buflen)
{
    char *ePtr;
    char WildXPR[64];
    char *p;
    int n;

    if ((ePtr = getenv ("XPR_PATH")) == NULL)
    {
	if (*LibDir)
	{
	    ePtr = LibDir;
	}
        else
	{
	    ePtr = "";
	}
    }

    strcat (strcpy (WildXPR, ePtr), "xpr*.library");

    n = getfnl (WildXPR, buf, (unsigned) buflen, 2L);
    return n;
}
/*
 * Use the XPR 'name'.
 * We use a private spawn() function here as c68 pipes are a little funny
 * and exec/fork wants 'C' file descriptors rather than QDOS ID, which, for
 * us, is a little inconvenient.
 */
 
XPRUse (pName)
char *pName;
{
    char ExeName[64];
    char *ePtr;
    int n, val = 0;
    SJOB s;
    short msgno;
    int xstat;

    memset (&s, '\0', sizeof (SJOB));

    if (XPRPid > 0)
    {
	XPRKill ();
    }

    if (pName && *pName)
    {
	OutError ("Loading XPR%s Library ....\n", pName);
        /*
         * Initialise the HostMon checking
         */
	abuf = rbuf;
	if ((ePtr = getenv ("XPR_PATH")) == NULL)
	{
	    if (*LibDir)
	    {
		ePtr = LibDir;
	    }
	    else
	    {
	        ePtr = "";
	    }
	}

	sprintf (ExeName, "%sxpr%s.library", ePtr, pName);

	OPipe = io_open ("pipe_8", 0);

	s.FileName = ExeName;           /* File we want */
	s.ComStr = NULL;                /* No command string */
	s.InFile = "pipe_0";            /* stdin */
	s.inflag = OPipe;               /* open type */
	s.OuFile = "pipe_8";            /* stdout */
	s.serr = pChId2;                /* stderr channel ID */
	s.Wait = 0;                     /* No wait (concurrent job) */
	s.Prio = 16;                    /* Priority */
	s.JobId = -1;                   /* Owned by me */

	if ((xstat = spawn (&s)) == 0)
	{
	    /*
	     * We get back the new job iD and the stdout pipe ID
	     * so we link the pipes up etc
	     */
	     
	    XPRPid = s.JobId;                   
	    IPipe = io_open ("pipe_0", s.ouid);

            /*
             * Set the startup message to the XPR
             */
             
	    msgno = XMSG_START;
	    io_sstrg (OPipe, -1, (char *) &msgno, 2);

            /*
             * It will acknowledge, returning its attributes
             * via the THING
             */
             
	    n = io_fstrg (IPipe, -1, (char *) &msgno, 2);
	    if (msgno == XMSG_START && n == 2)
	    {
		val = pXB->status;
	    }
	    else
	    {
		xstat = -1;
		OutError ("Incompatible XPR Library\n");
	    }
	}
	else
	{
	    OutError ("Can't load XPR Library (%d)\n", xstat);
	}
	if (xstat != 0)
	{
	    XPRPid = 0;
	    XPRKill ();
	    *MyProtocol = '\0';
	    val = 0;
	}
    }
    return val;
}
/*
 * User has selected 'Edit Prefs' or similar
 */

DoXPROptions ()
{
    int res, n;
    struct xpr_option *pxopt;
    int nopt;
    short msgno;

    /*
     * tell the XPR want we want
     */ 
    msgno = XMSG_SETUP;
    n = io_sstrg (OPipe, -1, (char *) &msgno, 2);

    /*
     * The XPR may send multiple XPR_SETDATA messages
     * if the setup can be in stages (i.e. KERMIT)
     *
     * We display the relevent setup data until we get the
     * XPR_STATUS message back, or a pipe error
     */
     
    if(PIPEOK(n))
    {
        n = io_fstrg (IPipe, -1, (char *) &msgno, 2);
        if(PIPEOK(n))
        {
            do
            {
	        if (msgno == XMSG_SETDATA)
	        {
	            pxopt = (struct xpr_option *) pXB->buf;
	            nopt = pXB->size;
                    /*
                     * Set the return value for the status
                     * and return it to the XPR
                     *
                     * We may then get another edit request !!
                     */
	            pXB->status = XPROptions (pxopt, nopt);
	            msgno = XMSG_STATUS;
	            io_sstrg (OPipe, -1, (char *) &msgno, 2);
	        }
	        n = io_fstrg (IPipe, -1, (char *) &msgno, (short) 2);
	        if (msgno == XMSG_STATUS)
	        {
	            res = pXB->status;
	        }
            }
            while (msgno != XMSG_STATUS && n > 0);
        }
    }
    return XPRTidy (res, n);
    
}

/*
 * Check for pipe errors (n <= 0)
 */
 
XPRTidy (int res, int n)
{
    if (n <= 0)
    {
	XPRKill ();
	*MyProtocol = '\0';
	OutError ("XPR Error on command (%d)\n", _oserr);
	return 0;
    }
    else
    {
	return res;
    }
}
/*
 * Do an XPR transfer
 */
XPRIO (char *f, int dirn)
{
    int msg;
    int flag;
    
    if (XPRPid > 0)
    {
	if (dirn == 's' || dirn == 'r')
	{
	    if (dirn == 's')
	    {
	        msg = XMSG_SEND;
	    }
	    else
	    {
	        msg = XMSG_RECV;
	    }
	    /*
	     * Give the XPR lots of CPU
	     */
	    mt_prior (-1, 1);
	    mt_prior (XPRPid, 32);
	    InitXWindow (f, IPipe, OPipe, msg, 0);
	}
    }
}

/*
 * Kill off the XPR
 */
 
XPRKill ()
{
    if (XPRPid > 0)
	mt_frjob (XPRPid, -1);
    if (IPipe > 0)
	io_close (IPipe);
    if (OPipe > 0)
	io_close (OPipe);
    IPipe = OPipe = XPRPid = 0;
}

/*
 * Check for an ZMODEM auto receive or send request
 * returns 0 for none, or the direction (SEND/RECEIVE) required
 */
 
XPRMon (ch)
char ch;
{
    char match, b;
    int k;
    static char Act = -1;
    int rval = 0;

    if (*abuf == 'x')
    {
	match = 0;
	if (ch >= '0' || ch <= '9')
	{
	    Act = ch;
	    match = 1;
	}
    }
    else
    {
	match = (ch == *abuf);
    }

    if (match)
    {
	abuf++;
	if (*abuf == '\0')
	{
	    while (io_fbyte(pS->pSerId, 2, &b) == 0)
		;       /* NULL loop, flush serial */

	    if (Act == '0')
	    {
	        rval  = XRECV;
	    }
            else if (Act > '0' && Act <= '9')
	    {
	        rval = XSEND;
	    }
	    
    	    abuf = rbuf;
	    Act = -1;
	}
    }
    else
    {
	abuf = rbuf;
	Act = -1;
	if (ch == *abuf)
	{
	    abuf++;
	}
    }
    return rval;
}

/*
 * Act on ZMODEM auto request
 */

AutoXPR(int dirn)
{	    
    if (dirn  == XRECV)
    {
        XPRIO (NULL, 'r');
    }
    else if (dirn == XSEND)
    {
        AutoSend ();
    }
}

/*
 * If the host has solicited an upload, we must get some files to send
 * or tell it to go away
 */

AutoSend ()
{
    int rval;
    char FileName[128];
    static char canit[] =
    {'*', '*', 24, 24, 24, 24, 24, 24, 24, 24, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 0};

    FileName[0] = '\0';
    OutError ("ZModem Auto-Send, NUL file to cancel\n");

    cputs("File Name : ", pChId2);
    io_fline (pChId2, -1, FileName, (short) strlen (FileName));

    if (*FileName > ' ')
    {
	OutError ("ZModem Auto-Send, SF5 to cancel\n");
	XPRIO (FileName, 's');
	rval = 1;
    }
    else
    {
	io_sstrg (pS->pSerId, -1, canit, (short) strlen(canit));
	rval = 0;
    }
    return rval;
}
