/*
 * xprthing.c - defines the "XPR Stuff" shared data thing used by
 *              the XPR system
 *
 *              requires HotStuff library
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <qdos.h>

#include <pd/thing.h>
#include <pd/xproto.h>

#include "xprdemo.h"
#include "xprmsg.h"

#define XTHG_NAME   "XPR Stuff";
#define XTHG_LEN    9
#define XPR_ID      0x302e3031

static char *Stuff = XTHG_NAME;
static char not_impl = 0;

PXPRTHG pXB = NULL;

PTHG_LINK find_thing(char *name)
{
    PTHG_LINK p;

    /*
     * Find a thing (without using THING system)
     */
     
    for (p = ((PTHG_LINK)(_sys_var + SYS_THGL))->nxtth; p; p = p->nxtth) 
    {
        if (XTHG_LEN == p->namelen &&
             strnicmp(Stuff, (char *)(p + 1), (unsigned) XTHG_LEN) == 0)
        {
            break;
        }
    }
    return p;
}

/*
 * Find and get usage of the XPR Stuff thing
 */
 
PTHG_LINK MakeThing(int jobid)
{
int size, m;
PTHG_LINK p;

    size = sizeof(THG_LINK) + ((XTHG_LEN + 1) & ~1);

    /*
     * Make one allocation for both linkage block and data
     */
        
    p = (PTHG_LINK) thg_malloc ((m = (size + sizeof(XPRTHG))), jobid);

    if(p)
    {
        p->namelen = XTHG_LEN;
        memcpy((p+1), Stuff, (unsigned) XTHG_LEN);
        p->verid = XPR_ID;

        /*
         * This is where the thing starts.
         */
         
        pXB = p->thing = ((char *)p + size);

        pXB->h.flag = THH_FLAG;
        pXB->h.type = THT_DATA;
    }
    return p;
}


GetXPRThing()
{
    int size;
    int sts;
    PTHG_LINK p = NULL;

    sts = getthg();

    /*
     * If we have the THING system, we allocate the linkage block
     * owned by this job, so if we get RJob'ed, QDOS will clean up
     * and remove it for us.
     */
     
    if(sts == 0)
    {
        while(sms_uthg(Stuff, -1, -1, NULL, &(void *)p, &size, & (void *) pXB))
        {
            p = MakeThing(-1);
            if(p == NULL || sms_lthg(p))
            {
                break;
	    }
	}
    }
    else
    {
        not_impl =1;
        p = find_thing(Stuff);
        if(p == NULL)
        {
            /*
             * If we don't have the THING system, we allocate the linkage block
             * owned by the system (job 0), so if we get RJob'ed, the THING
             * stays and the 'thing' list is still valid. 
             */
            if((p = MakeThing(0)))
            {
                PTHG_LINK q;

                q = (PTHG_LINK)(_sys_var + SYS_THGL);
                p->nxtth = q->nxtth;
                q->nxtth = p;
            }
        }
        else
        {
            pXB = p->thing;
        }
    }

    if(p && pXB)
    {
        /*
         * if we have a THING (pointer pXB), then we load up
         * the data areas
         */
         
        pXB->psaddr = pS;           /* Serial status block */
	pXB->chid = 0;              /* Status channel, none yet */
	pXB->job = 0;               /* Our job ID, only required if we want
	                               to be picked and waked by the PE */  
	pXB->logid = 0;             /* Log channel, none yet */
	pXB->check = 0;             /* Abort flag, Not yet ! */
        pXB->libdir = LibDir;       /* Were the _pref files are */
        pXB->xprname = MyProtocol;  /* The protocol we want */          
    }
    else
    {
        thg_free(p);
        p = NULL;
        pXB = NULL;
    }
}

/*
 * Free the Thing at exit
 */

FreeXPRThing()
{
    if(not_impl)
    {
        /*
         * No HOTKEY II (THING) system, so just adjust system variables
         * Thing list.
         */
         
        PTHG_LINK p,q;

        q =  (PTHG_LINK)(_sys_var + SYS_THGL);
        
        for (p = q->nxtth; p; p = p->nxtth) 
        {
            if (p->namelen == XTHG_LEN &&
                        strnicmp(Stuff, (char *)(p + 1), (unsigned) XTHG_LEN) == 0)
            {
                q->nxtth = p->nxtth;
                /*
                 * Here we can free the linkage and data areas
                 */
                thg_free(p);
                break;
            }
            q = p;
        }
    }
    else
    {
        /*
         * We have HOTKEY II (THING) system, so do it correctly
         */
        if(XPRPid)
        {
            /*
             * If the XPR is still up, free the thing for it first
             */
            sms_fthg(Stuff, XPRPid, 0, NULL, NULL, NULL);
        }
        /*
         * Now free and remove the THING
         */
        sms_fthg(Stuff, -1, 0, NULL, NULL, NULL);
        sms_rthg(Stuff);
    }
}
