#include <stdio.h>
#include <string.h>
#include "xproto.h"
#include "xprsealink.h"

static char *blkchk[] =
{"Checksum", "CRC-16"};
static char *timefmt = "%02ld:%02ld:%02ld";

long XProtocolCleanup (X)
PXIO X;
{
    if (X->xpr_data)
    {
	free (X->xpr_data);
	X->xpr_data = NULL;
    }
    return (1);
}

DOUpdate (X, flag)
PXIO X;
int flag;
{
    XPRU u;
    int secs, elapsed;
    char buff1[20], buff2[20];
    char text[64];
    
    if (X->xpr_update)
    {
        setmem(&u, sizeof(XPRU), '\0');

	time (&secs);

	secs -= XD->starttime;

	u.xpru_updatemask = XPRU_ELAPSEDTIME;

	sprintf (buff1, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);

	u.xpru_elapsedtime = buff1;

	if (secs && XD->bytes)
	{
	    u.xpru_datarate = XD->bytes / secs;
	    u.xpru_updatemask |= XPRU_DATARATE;
	}
	else
	{
	    u.xpru_datarate = 0;
	}
	/*
	 * If we know the file size and the data rate we can compute the
	 * estimated time.
	 */

	if (XD->filesize)
	{
	    if (u.xpru_datarate)
	    {
		u.xpru_updatemask |= XPRU_EXPECTTIME;
		secs = XD->filesize / u.xpru_datarate;
		sprintf (buff2, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);
		u.xpru_expecttime = buff2;
	    }
	}

	switch (flag)
	{
	    case MODE:
		u.xpru_blockcheck = XD->chktec ? blkchk[1] : blkchk[0];
		u.xpru_timeouts = XD->numtime;
		u.xpru_updatemask |= XPRU_TIMEOUTS | XPRU_BLOCKCHECK;
		break;
	    case TIMEOUT:
		u.xpru_blockcheck = XD->chktec ? blkchk[1] : blkchk[0];
		u.xpru_errormsg = "Timeout";
		u.xpru_timeouts = XD->numtime;
		u.xpru_updatemask |= XPRU_TIMEOUTS | XPRU_ERRORMSG | XPRU_BLOCKCHECK;
		break;
	    case LASTBLK:
		u.xpru_updatemask |= XPRU_MSG;
		u.xpru_msg = "Send NULL block";
		break;
	    case DONE:
		u.xpru_updatemask |= XPRU_MSG;
		u.xpru_msg = "Done";
		break;
	    case OK:
		u.xpru_updatemask |= XPRU_BLOCKS | XPRU_BLOCKSIZE | XPRU_BLOCKCHECK | XPRU_BYTES;
                u.xpru_bytes = XD->bytes;
		u.xpru_blocks = XD->blksnt;
		u.xpru_blocksize = 128;
		u.xpru_blockcheck = XD->chktec ? blkchk[1] : blkchk[0];
		break;
	    case NAK:
	        u.xpru_errors = XD->numerrs;
	        u.xpru_errormsg = "Received NAK for sector";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
                break;
            case FAILED:
	        u.xpru_errors = XD->numerrs;
	        u.xpru_msg = "Failed";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_MSG;
                break;
            case WCEOT:
	        u.xpru_errors = XD->numerrs;
	        u.xpru_errormsg = "No ACK for EOT";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
                break;
	    case CRCERR:
            case CHKERR:    
	        u.xpru_errors = XD->numerrs;
	        u.xpru_errormsg = text;
                sprintf(text, "%s error at block %d (%d)",
                    (flag==CRCERR) ? "CRC" : "Checksum", XD->blksnt, XD->bytes); 
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
                break;
            case FUTURE:
		u.xpru_updatemask |= XPRU_ERRORMSG;
                u.xpru_errormsg = "Future block";
                break;
            case NOFILES:
		u.xpru_updatemask |= XPRU_ERRORMSG;
                u.xpru_errormsg = "No files selected";
                break;
            case ABORTED:
		u.xpru_updatemask |= XPRU_MSG;
                u.xpru_msg = "Transfer Aborted";
                break;
            case FILERR:
		u.xpru_updatemask |= XPRU_ERRORMSG;
                u.xpru_msg = "File ERROR";
                break;
            case DUPSEC:
		u.xpru_errormsg = text;
                sprintf(text, "Duplicate Sector %d", XD->blksnt);
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	}
	(*X->xpr_update) (&u);
    }
}
