#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "xproto.h"
#include "xprsealink.h"

#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};

char _prog_name[] = "<QDOS XPR-SEALink v2.30>";
static QVS (_qcfx_, "QCFX>>01\x00\x0aXPRSEALINK\x00\x042.30\xFF\xFF");

char *find_option (char *, char);

static struct xpr_option op[] =
{
    {"SEALink options:", XPRO_HEADER, NULL, 0},
    {"QL Translate (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"Default receive path:", XPRO_STRING, NULL, 0}
};

static char option_q[4], option_p[32];

CheckSetup (X)
PXIO X;
{

    if (X->xpr_data == NULL)
    {
         if ((X->xpr_data = (PXPD) malloc (sizeof (XPD))) == NULL)
         {
             ioerr (X, "Not enough memory");
             return ERROR;
         }
         setmem (X->xpr_data, sizeof (XPD), '\0');
         XD->QLXlate = 'N';
#ifdef QDOS
         strcpy(XD->DefDir, "ram2_");
#else
         strcpy(XD->DefDir, "t:");
#endif
    }

     
    return OK;
}
Welcome(X)
PXIO X;
{
XPRU u;

    setmem(&u, sizeof(XPRU), '\0');
    u.xpru_protocol = "SEALink";
    u.xpru_msg = "Waiting ... ";
    u.xpru_updatemask = XPRU_PROTOCOL | XPRU_FILESIZE |
                  XPRU_MSG | XPRU_BLOCKS | XPRU_BLOCKSIZE |
                  XPRU_BYTES | XPRU_ERRORS | XPRU_TIMEOUTS |
                  XPRU_PACKETDELAY | XPRU_CHARDELAY | XPRU_DATARATE;
    (*X->xpr_update) (&u);

    /*
     * AND ZERO VITAL DATA ????
     */

}
/**********************************************************
 **********************************************************/

long XProtocolSetup (X)
PXIO X;
{

    char buf[80], *p;
    int i, len;

    if (CheckSetup (X) == ERROR)
    {
         return XPRS_FAILURE;
    }

    /* If options string passed by comm prog, use it; else prompt user */
    if (X->xpr_filename)
    {
         strcpy (buf, X->xpr_filename);
    }
    else
    {
         /* If xpr_options() implemented by comm program, use it */
         if (X->xpr_extension >= 1 && X->xpr_options)
         {
             op[1].xpro_value = option_q;
             op[1].xpro_length = sizeof (option_q);
             *option_q = XD->QLXlate;
             *(option_q + 1) = '\0';
             op[2].xpro_value = option_p;
             op[2].xpro_length = sizeof (option_p);
             strcpy(option_p, XD->DefDir);

             /* Convert Y/N used elsewhere into "yes"/"no" required by spec */
             strcpy (op[1].xpro_value, (op[1].xpro_value[0] == 'Y') ? "yes" : "no");

             (*X->xpr_options) (op, 3);

             /* Convert "yes"/"no" or "on"/"off" into Y/N */
         if(op[1].xpro_value[0] == 'Y' ||
            op[1].xpro_value[0] == 'y' ||
            stricmp (op[1].xpro_value, "yes") == 0 ||
            stricmp (op[1].xpro_value, "on") == 0)
         {
                  op[1].xpro_value[0] = 'Y';
         }
         else
         {
                  op[1].xpro_value[0] = 'N';
         }


            sprintf (buf, "Q%c,P%s", *option_q, option_p);
         }
         else
         {
             /* Start buffer with current settings so user can see/edit them. */
             sprintf(buf, "Q%c,P%s",
                        XD->QLXlate, XD->DefDir);
             if (X->xpr_gets)
                  (*X->xpr_gets) ("SEALink options:", buf);
         }
    }
    /* Upshift options string for easier parsing */
    strupr (buf);

    /*
     * Merge new T(ext) option into current settings if given "TY" = Force
     * Text mode on, "TN" = Force Text mode off, "T?" = Use other end's text
     * mode suggestion (default to binary) "TC" = Ask Comm program for file
     * type
     */

    if (p = find_option (buf, 'Q'))
    {
         if (*p == 'Y' || *p == 'N')
         {
             XD->QLXlate  = *p;
             XD->Xmode = (XD->QLXlate == 'Y');
         }
         else
             ioerr (X, "Invalid Q flag ignored; should be Y or N");
    }

    if (p = find_option (buf, 'P'))
    {
        strcpy(XD->DefDir, p);
    }

    return XPRS_SUCCESS | XPRS_NORECREQ;
}


XSetup (X)
PXIO X;
{
    if (!X->xpr_update)
    {
         return ERROR;
    }
    if (!X->xpr_fopen || !X->xpr_fclose || !X->xpr_fread
         || !X->xpr_fwrite || !X->xpr_fseek || !X->xpr_sread
         || !X->xpr_swrite)
    {
         ioerr (X, "Comm prog missing required function(s); see docs");
         return ERROR;
    }

    if (CheckSetup (X) == ERROR)
    {
         ioerr (X, "Setup Fails");
         return ERROR;
    }

    XD->starttime = 0;
    XD->filesize = 0;
    XD->blksnt = 0;
    XD->bytes = 0;
    XD->numerrs = 0;
    XD->numtime = 0;
    XD->abort = 0;
    XD->chktec = 0;
    return OK;

}

char *find_option (buf, option)
char *buf;
char option;
{
    while (*buf)
    {
        buf += strspn(buf," ,\t\r\n");
        if (*buf == option)
        {
            return ++buf;
        }
        buf += strcspn(buf," ,\t\r\n");
    }
    return NULL;
}

ioerr (X, text)
PXIO X;
char *text;
{
    XPRU u;
    if(X->xpr_update)
    {
         setmem(&u, sizeof(XPRU), '\0');
         u.xpru_msg = text;
         u.xpru_updatemask = XPRU_MSG;
         (*X->xpr_update)(&u);
    }
}

