#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <qdos.h>

#include <pd/thing.h>
#include <pd/xproto.h>

#include "psblk.h"
#include "xprmsg.h"

#define XTHG_NAME   "XPR Stuff";
#define XTHG_LEN    9

static char *Stuff = XTHG_NAME;

long (*_cmdchannels) () = NULL;
void (*_consetup) () = NULL;
long (*_conread) () = NULL;
long (*_conwrite) () = NULL;
char *_endmsg = NULL;

#define _STDIO_VA_  ,

#define FILBUF  4096

static char *maxcpy (char *, char *, int);
extern char *_sys_var;

static char *XPR_ERROR = "*** XPR ERROR ***";
static short XPRERRLEN = sizeof(XPR_ERROR);

static PSERBLK pS = NULL;
static PXPRTHG pXB = NULL;

static char XFileName[256];
static char Prefs[256];
static int IPipe, OPipe;

static char sx_proto[32];
static char sx_filename[50];
static char sx_msg[50];
static char sx_errmsg[50];
static char sx_bcheck[8];
static char sx_expect[16];
static char sx_elapsed[16];

/*
 * This is our ghost xpr_update structure to keep track of what is happening
 * so we can re-draw the window if we're a PTR program that gets picked when
 * iconified.
 */
 
static struct XPR_UPDATE sx =
{
    0,
    sx_proto,
    sx_filename,
    0,
    sx_msg,
    sx_errmsg,
    0, 0, 0, 0, 0, 0, 0, 0,
    sx_bcheck,
    sx_expect,
    sx_elapsed,
    0
};

static short sx_dirn = 0;
static short IOFlag = 0;

/* This structure is used by our internal file buffering
 * Not using stdio saves about 4 Kbytes per protocol
 */

#define BUFSIZE 16*1024

char * FullName(char *file, char *myname);

typedef struct {
    int Filebufpos;
    int Filebuflen;
    int Filebufcnt;
    int Filebufmax;
    char * Filebuf;
    char * Filebufptr;
    int fptr;
    short Fileflush;
} FCTL, *PFCTL;

/*
 * Allocate a file buffer, try 16K buffer (write), 32k (read), then reduce it
 * Give up if can't get 1K.
 */

char * falloc(PFCTL v, char mode)
{
    v->Filebufmax = BUFSIZE;

    if(mode != 'r')
    {
        v->Filebufmax = v->Filebufmax << 1;
    }
    for (; v->Filebufmax >= 1024; v->Filebufmax -= 1024)
    {
        if((v->Filebuf = malloc((size_t)v->Filebufmax)) != NULL)
        {
            break;
        }
    }
    return v->Filebuf;
}

/*
 * Open a buffered file
 */
 
static PFCTL  bfopen (char * name, char *mode)
{
    PFCTL v = NULL;
    int iomode;
    short ifree = 0;
    char *p;
    char myname[64];

    if((v = malloc((size_t)sizeof(FCTL))))
    {
        p = name;
        if(falloc(v, *mode))
        {
            v->Filebufpos = v->Filebuflen = v->Filebufcnt = 0;
            v->Fileflush = FALSE;
            v->Filebufptr = v->Filebuf;
            
            if (*mode == 'a')
            {
                iomode = 0;
            }
            else if(*mode == 'r')
            {
                iomode = 1;
            }
            else
            {
                iomode = 3;
            }
            
            p = FullName(name, myname);
            
            v->fptr = io_open(p, iomode);

            if(v->fptr > 0)
            {
                if(iomode == 0)
                {
                    struct qdirect s;
                    fs_headr(v->fptr, -1, (char *)&s, (int) sizeof(s));	    
                    fs_pos (v->fptr, s.d_length, 0);
                }
            }
            else
            {
                ifree = 1;
                free(v->Filebuf);
            }
        }
        else
        {
            ifree = 1;
        }
        if(ifree)
        {
            free(v);
            v = NULL;
        }
    }
    return v;
}

/*
 * Close a buffered file
 */
 
static void bfclose (PFCTL v)
{
    if (v && v->fptr > 0)
    {
	if (v->Fileflush)
	{
            fs_save(v->fptr, v->Filebuf, v->Filebufcnt);
	}
	io_close (v->fptr);
        free(v->Filebuf);
        free(v);
        v = NULL;
    }
}
				/* End of void bfclose() */
/*
 * Seek in a buffered file
 */

static void bfseek (PFCTL v, long pos, int x)
{
    long offset;
    /*
     * If new file position is within currently buffered section, reset
     * pointers
     */

    if (pos >= v->Filebufpos && pos < v->Filebufpos + v->Filebuflen)
    {
	offset = pos - v->Filebufpos;
	v->Filebufptr = v->Filebuf + offset;
	v->Filebufcnt = v->Filebuflen - offset;
	/*
	 * Otherwise, fseek() file & discard buffer contents to force new
	 * read
	 */
    }
    else
    {
	fs_pos(v->fptr, pos, 0L);
	v->Filebuflen = v->Filebufcnt = 0;
	v->Filebufpos = pos;
    }
}
				/* End of void bfseek() */
/*
 * Read a buffered file
 */

static long  bfread (char *buf, int elmsize, int elms, PFCTL v)
{
    long count, total = 0;
    int length;

    length = elmsize * elms;
    
    /* Keep going until entire request completed */
    while (length > 0)
    {
	/* Copy as much of the request as possible from the buffer */
	count = (length <= v->Filebufcnt) ? length : v->Filebufcnt;
	memcpy (buf, v->Filebufptr, (unsigned) count);
	buf += count;
	total += count;
	length -= count;
	v->Filebufptr += count;
	v->Filebufcnt -= count;

	/* If we've emptied the buffer, read next buffer's worth */

	if (!v->Filebufcnt)
	{
	    v->Filebufpos += v->Filebuflen;
	    v->Filebufptr = v->Filebuf;

	    v->Filebufcnt = v->Filebuflen
		= io_fstrg (v->fptr, -1, v->Filebuf, (short) v->Filebufmax);

	    /* If we hit the EOF, return with however much we read so far */

	    if (!v->Filebufcnt)
		break;
	}
    }
    return total;
}				/* End of long bfread() */
				/* End of void bfseek() */
/*
 * Write a buffered file
 */

static long  bfwrite (char *buf, int elmsize, int elms, PFCTL v)
{
    long count, total = 0;
    int length;

    length = elmsize * elms;

    /* Keep going until entire request completed */
    while (length > 0)
    {
	/* Copy as much as will fit into the buffer */
	count = v->Filebufmax - v->Filebufcnt;
	if (length < count)
	    count = length;
	memcpy (v->Filebufptr, buf, (unsigned) count);
	buf += count;
	total += count;
	length -= count;
	v->Filebufptr += count;
	v->Filebufcnt += count;
	v->Fileflush = TRUE;

	/* If we've filled the buffer, write it out */
	if (v->Filebufcnt == v->Filebufmax)
	{
	    fs_save (v->fptr, v->Filebuf, v->Filebufcnt);
	    v->Filebufptr = v->Filebuf;
	    v->Filebufcnt = 0;
	    v->Fileflush = FALSE;
	}
    }
    return total;
}

static char * FullName(char *name, char *myname)
{
    char *p;

    if(isdirdev (name) == 0)
    {
        if(getcwd(myname, 32))
        {
            p = strcat(myname, name);
        }
    }
    else
    {
        p = name;
    }
    return p;
}

/*
 * See if we have any serial data waiting
 */
 
CheckQueue()
{
    if(IOFlag && pS->ReadQue)
    {
        return (pS->Count > 0);
    }
    else
    {
        return (io_pend(pS->pSerId, 0) == 0);
    }
}

/*
 * Initialise our ghost update structure
 */

static InitSX ()
{
    *sx_filename = 0;
    *sx_msg = 0;
    *sx_errmsg = 0;
    *sx_bcheck = 0;
    *sx_expect = 0;
    *sx_elapsed = 0;
    sx.xpru_updatemask = XPRU_PROTOCOL;
}

/*
 * XPR Options processing
 */
 
static int XOptions (struct xpr_option *Opts, int n)
{
    short msgno;

    msgno = XMSG_SETDATA;

    pXB->size = n;
    pXB->buf = Opts;

    io_sstrg (OPipe, -1, (char *) &msgno, 2);
    /*
     * The comms program does its work here
     */
    io_fstrg (IPipe, -1, (char *) &msgno, 2);
    return pXB->status;
}

/*
 * Set serial parameters
 */
 
static int XSetSer (int val)
{
    char bx;

    /*
     * QTPI & QeM turn off any special XON/XOFF procesing (1)
     * or CompuServe 7 bit mode (2)
     */
     
    if ((val & 1))
    {
	pS->bFlow = 0;
    }
    if((val & 2))
    {
        pS->QCIS = 0;
    }
    bx = *(_sys_var + 0xA0);
    return (int) (bx & 0x7);
}

/*
 * Get file info
 */

static Xfinfo (filnam, flag)
char *filnam;
int flag;
{
    int rc = 0;
    char myname[64];
    char *p;
    int fd = 0;

    if (filnam && *filnam)
    {
        p = FullName(filnam, myname);
        fd = io_open(p, 1);
        if (fd >  0)
	{
            struct qdirect s;
            fs_headr(fd, -1, (char *)&s, (int) sizeof(s));	    
            io_close(fd);
            switch (flag)
            {
	        case 1:
	            rc = s.d_length;
                    break;
	        case 2:
	            if (s.d_type == 0)
		    {
		        rc = 2;
		    }
		    else
		    {
		        rc = 1;
		    }
		    break;
	    }
        }
    }        
    return rc;
}

/*
 * Flush serial input data
 */

static Xsflush ()
{
    char ch;

    while (SRead (&ch, 1, 2) == 1)
        ;
}

/*
 * One Xgets
 */

static Xgets (prompt, buf)
char *prompt;
char *buf;
{
    strcpy (Prefs, buf);
}
/*
 * Another
 */
static MyGets (prompt, buf)
char *prompt;
char *buf;
{
    return 0;
}

/*
 * Check comms program abort flag
 */

static Xchkabort ()
{
    return pXB->check;
}

/*
 * Update status display, defined by pXB->flag
 *  0 - Do normal screen update
 *  1 - Don't do updates
 *  2 - Refresh all display
 */
 
static Update (msg, x, y, len)
char *msg;
int x, y, len;
{

    int i;
    short ml;

    if (pXB->flag == 0)
    {
	ml = strlen (msg);
	if (ml > len)
	{
	    *(msg + len) = 0;
	    *(msg + len - 1) = '';
	    ml = len;
	}

	if (pXB->chid > 0)
	{
	    sd_pos (pXB->chid, -1, x, y);
	    io_sstrg (pXB->chid, -1, msg, ml);

	    for (i = ml; i < len; i++)
	    {
		io_sbyte (pXB->chid, -1, ' ');
	    }
	}
	else
	{
	    int chn = getchid (2);
	    io_sstrg (chn, -1, XPR_ERROR, XPRERRLEN);
	    io_sstrg (chn, -1, msg, ml);
	}
    }
    else if (pXB->flag == 2)
    {
	pXB->flag = 0;
	Xupdate (&sx);
    }
}

/*
 * The XPR update call-back. Note we only do some things if we are really
 * updating and not just redrawing the screen. We also have to do some funny
 * stuff to ensure that the percentage display has the right background
 * colour.
 */
 
static Xupdate (u)
struct XPR_UPDATE *u;
{
    char txt[16];
    int i;
    static int lb;
    static int fs;
    static short blk[] = {438, 10, 0, 110};
    short umask;

    umask = (u != &sx);

    if (pXB->chid)
    {
	if (u->xpru_updatemask & XPRU_PROTOCOL)
	{
	    Update (u->xpru_protocol, 20, 0, 50);
	}
	if (u->xpru_updatemask & XPRU_FILENAME)
	{
	    if (u->xpru_filename && u->xpru_filename[0])
	    {
		Update (u->xpru_filename, 20, 1, 40);
                if(umask)
                {
                    if(strcmp (u->xpru_filename, sx.xpru_filename) != 0)
		    {
		        if(pXB->logid > 0)
		        {
                            /*
                             * Log the previous file
                             */
                             
			    LogInfo ();
			}
			InitSX ();
			sx.xpru_filesize = sx.xpru_datarate = 0;
		    }
		    maxcpy (sx.xpru_filename, u->xpru_filename, 39);
		}
	    }
	}
	if (u->xpru_updatemask & XPRU_FILESIZE)
	{
	    if (umask)
	    {
		sx.xpru_filesize = u->xpru_filesize;
	    }
	    sprintf (txt, "%-8d", (fs = u->xpru_filesize));
	    Update (txt, 20, 2, 8);
	    blk[0] = 438;
	    blk[2] = 0;
	    if (pXB->flag == 0)
	    {
		sd_fill (pXB->chid, -1, 0, (struct QLRECT *) blk);
	    }
	    lb = 0;
	}
	if (u->xpru_updatemask & XPRU_BYTES)
	{
	    if (umask)
	    {
		sx.xpru_bytes = u->xpru_bytes;
	    }
	    sprintf (txt, "%-8d", u->xpru_bytes);
	    Update (txt, 20, 3, 8);
	    if (fs > 0 && pXB->flag == 0)
	    {
		int nb, pct, j;
		char *p, temp[8];

		if (u->xpru_bytes >= fs)
		{
		    nb = 438;
		}
		else
		{
		    nb = 438 * u->xpru_bytes / fs;
		}
		blk[0] = nb - lb;
		blk[2] = lb;
		lb = nb;
		pct = nb * 100 / 438;
		sprintf (temp, "%3d%%", pct);
		sd_fill (pXB->chid, -1, 2, (struct QLRECT *) blk);
		sd_pos (pXB->chid, -1, 35, 11);

		for (j = 36, p = temp; *p; p++, j++)
		{
		    if (6 * j > nb)
		    {
			sd_setst (pXB->chid, -1, 0);
		    }
		    else
		    {
			sd_setst (pXB->chid, -1, 2);
		    }
		    io_sbyte (pXB->chid, -1, (*p));
		}
		sd_setst (pXB->chid, -1, 0);
	    }
	}
	if (u->xpru_updatemask & XPRU_BLOCKCHECK)
	{
	    if (umask)
	    {
		strcpy (sx.xpru_blockcheck, u->xpru_blockcheck);
	    }
	    Update (u->xpru_blockcheck, 20, 4, (short) strlen (u->xpru_blockcheck));
	}
	if (u->xpru_updatemask & XPRU_EXPECTTIME)
	{
	    if (umask)
	    {
		maxcpy (sx.xpru_expecttime, u->xpru_expecttime, 15);
	    }
	    Update (u->xpru_expecttime, 20, 5, 12);
	}
	if (u->xpru_updatemask & XPRU_MSG)
	{
	    if (umask)
	    {
		maxcpy (sx.xpru_msg, u->xpru_msg, 49);
	    }
	    Update (u->xpru_msg, 20, 6, 50);
	}
	if (u->xpru_updatemask & XPRU_ERRORMSG)
	{
	    if (umask)
	    {
		maxcpy (sx.xpru_errormsg, u->xpru_errormsg, 49);
	    }
	    Update (u->xpru_errormsg, 20, 7, 50);
	}
	if (u->xpru_updatemask & XPRU_ERRORS)
	{
	    if (umask)
	    {
		sx.xpru_errors = u->xpru_errors;
	    }
	    sprintf (txt, "%-4d", u->xpru_errors);
	    Update (txt, 20, 8, 4);
	}
	if (u->xpru_updatemask & XPRU_PACKETTYPE)
	{
	    if (umask)
	    {
		sx.xpru_packettype = u->xpru_packettype;
	    }
	    sprintf (txt, "'%c'", u->xpru_packettype);
	    Update (txt, 20, 9, 3);

	}
	if (u->xpru_updatemask & XPRU_DATARATE)
	{
	    if (umask)
	    {
		sx.xpru_datarate = u->xpru_datarate;
	    }
	    sprintf (txt, "%-4d", u->xpru_datarate);
	    Update (txt, 20, 10, 4);
	}
	if (u->xpru_updatemask & XPRU_BLOCKS)
	{
	    if (umask)
	    {
		sx.xpru_blocks = u->xpru_blocks;
	    }
	    sprintf (txt, "%-4d", u->xpru_blocks);
	    Update (txt, 60, 3, 4);
	}
	if (u->xpru_updatemask & XPRU_BLOCKSIZE)
	{
	    if (umask)
	    {
		sx.xpru_blocksize = u->xpru_blocksize;
	    }
	    sprintf (txt, "%-4d", u->xpru_blocksize);
	    Update (txt, 60, 4, 4);
	}
	if (u->xpru_updatemask & XPRU_ELAPSEDTIME)
	{
	    if (umask)
	    {
		maxcpy (sx.xpru_elapsedtime, u->xpru_elapsedtime, 15);
	    }
	    Update (u->xpru_elapsedtime, 60, 5, 12);
	}
	if (u->xpru_updatemask & XPRU_TIMEOUTS)
	{
	    if (umask)
	    {
		sx.xpru_timeouts = u->xpru_timeouts;
	    }
	    sprintf (txt, "%-4d", u->xpru_timeouts);
	    Update (txt, 60, 8, 4);
	}
	if (u->xpru_updatemask & XPRU_PACKETDELAY)
	{
	    if (umask)
	    {
		sx.xpru_packetdelay = u->xpru_packetdelay;
	    }
	    sprintf (txt, "%-4d", u->xpru_packetdelay);
	    Update (txt, 60, 9, 4);
	}
	if (u->xpru_updatemask & XPRU_CHARDELAY)
	{
	    if (umask)
	    {
		sx.xpru_chardelay = u->xpru_chardelay;
	    }
	    sprintf (txt, "%-4d", u->xpru_chardelay);
	    Update (txt, 60, 10, 4);
	}
	if (umask)
	{
	    sx.xpru_updatemask |= u->xpru_updatemask;
	}
    }
}

/*
 * File list processing, extracts file names from
 * string of files, wildcards, @file etc and combinations
 * thereof
 */

static char *FList = NULL, *FPtr;

static char *ExtractFile (char *pat, int *pflag)
{
    static int fchid = 0;
    char *p, *q;
    static char text[64];

    if (fchid > 0)
    {
        int n;
        n = io_fline(fchid, -1, text, 64);
	if (n > 0)
	{
	    if ((p = strchr ((q = text), '\n')))
	    {
		*p = '\0';
	    }
	}
	else
	{
	    io_close (fchid);
	    fchid = 0;
            /*
             * Recursion: see recursion etc
             */
	    q = ExtractFile (pat, pflag);
	}
    }
    else
    {
	q = strtok ((*pflag) ? NULL : pat, " ,\t");
	(*pflag)++;
	if (q && *q == '@')
	{
	    fchid = io_open (q + 1, 0);
	    q = ExtractFile (pat, pflag);
	}
    }
    return q;
}

static char *MakeFileList (char *pat)
{
    char this[128];

    static char *wild = "*?!";

    char *p, *q;
    int n, totl = 0;
    int FSize;
    int Done, k;

    FSize = FILBUF;
    if (FList == NULL)
    {
	FList = FPtr = (char *) malloc ((size_t) FSize);
    }
    else
    {
	FPtr = FList;
    }
    if (FList)
    {
	for (Done = k = 0; !Done;)
	{
	    if ((p = ExtractFile (pat, &k)) == NULL)
	    {
		Done = 1;
		continue;
	    }
	    if (strpbrk (p, "*?!") == NULL)
	    {
		strcpy (FPtr, p);
		FPtr = strchr (FPtr, '\0');
		*++FPtr = '\0';
		FSize -= (FPtr - FList);
		totl++;
	    }
	    else
	    {
		n = getfnl (p, FPtr, (unsigned int) FSize, 0);
		if (n > 0)
		{
		    totl += n;
		    for (; *FPtr; FPtr = strchr (FPtr, '\0') + 1)
			;	/* NULL Loop to find the end */
		    FSize -= (FPtr - FList);
		}
	    }
	}
	if (totl == 0)
	{
	    free (FList);
	    FList = NULL;
	}
    }
    return FList;
}

/*
 * The actual XPR first/next routines
 */

static ffirst (char *buf, char *pat)
{
    FPtr = FList;
    return fnext (0, buf, pat);
}

static fnext (int flag, char *buf, char *pat)
{
    int n;
    char *p;

    if (FPtr && *FPtr)
    {
	strcpy (buf, FPtr);
	FPtr += strlen (FPtr) + 1;
	return 1;
    }
    else
    {
	free (FList);
	FList = NULL;
	return 0;
    }
}

/*
 * The Serial read routine
 * uses either QeM/QTPI embedded serial handler
 * or straight QDOS calls
 */
 
static SRead (buf, nchar, timeout)
char *buf;
int nchar;
int timeout;
{
    int m;

    if(IOFlag && pS->ReadQue)
    {
        int n;

        m = 0;
        n = (*pS->ReadQue) (timeout * 5, buf, nchar, &m);
        if (n != 0)
        {
	    m = 0;
        }
    }
    else
    {
        m = io_fstrg(pS->pSerId, (timeout * 5), buf, nchar);
        if(m < 0)
        {
            m = 0;
        }
    }
   return m;
}

/*
 * The Serial write routine
 * uses either QeM/QTPI embedded serial handler
 * or straight QDOS calls
 */
 
static SWrite (buf, len)
char *buf;
int len;
{
    if(IOFlag && pS->SendChr)
    {
        int n;
        char *p;
        for (p = buf, n = len; n; n--, p++)
        {
	    (*pS->SendChr) (*p);
        }
    }
    else
    {
        io_sstrg(pS->pSerId, -1, buf, len);
    }
    return 0;
}

/*
 * The easiest way to debug this is to have the XPR stderr as a channel owned
 * by the parent job, then we can see what's going on.
 *
 * Just in case you need it :-)
 */
#ifdef REPORT
Report (char *ctl _STDIO_VA_)
{
    va_list va;
    char t[80];
    int xc;

    if((xc = getchid (2)) > 0)
    {
        va_start (va, ctl);
        vsprintf (t, ctl, va);
        va_end (va);

        io_sstrg (xc, -1, t, (short) strlen (t));
    }
}
#endif

static int sptr = 0;

/*
 * Some initialisation stuff
 * returns a pointer to the file list
 */

static char *GetXParams ()
{
    char *p, *s;

    sptr = pXB->check;
    pXB->check = 0;

    InitSX ();

    p = (char *) pXB->buf;
    if (p && *p)
    {
	s = MakeFileList (p);
    }
    else
    {
	s = p;
    }
    return s;
}

/*
 * Gets a preference file if possible, returns the QDOS channel ID
 * for the file.
 */

static int GetPrefsFile(char *PrefFile, char *xProt, char *pdev, int mode)
{
    char *sp;
    int fchid;
    
    if ((sp = getenv ("XPR_PREFS")))
    {
        sprintf (PrefFile, "%s%s_Prefs", sp, xProt);
        if((fchid = io_open (PrefFile, mode)) > 0)
        {
            return fchid;
        }
    }
    if(pdev && *pdev)
    {
        sprintf (PrefFile, "%s%s_Prefs", pdev, xProt);
        if((fchid = io_open (PrefFile, mode)) > 0)
        {    
            return fchid;
        }
    }
    *PrefFile = '\0';
    return 0;
}

PTHG_LINK find_thing(char *name)
{
    PTHG_LINK p;

    for (p = ((PTHG_LINK)(_sys_var + SYS_THGL))->nxtth; p; p = p->nxtth) 
    {
        if (XTHG_LEN == p->namelen &&
             strnicmp(Stuff, (char *)(p + 1), (unsigned) XTHG_LEN) == 0)
        {
            break;
        }
    }
    return p;
}

static GetXPRThing()
{
    int sts;
    PTHG_LINK p; 
    int vers;

    sts = sms_uthg(Stuff, -1, -1, NULL, &(void *)p, &vers, &(void *)pXB);
    if(sts == ERR_NI)
    {
        if(p = find_thing(Stuff))
        {
            pXB = p->thing;
        }
    }
    else if(sts)
    {
        pXB = NULL;
    }
}
static BombOut()
{
    io_sstrg(getchid(2), -1, XPR_ERROR, XPRERRLEN);
    exit (-1);
}    
main (int ac, char **av)
{
    static struct XPR_IO X =
    {
	NULL,
	bfopen,
	bfclose,
	bfread,
	bfwrite,
	SRead,
	SWrite,
	Xsflush,
	Xupdate,
	Xchkabort,
	NULL,
	MyGets,
	XSetSer,
	ffirst,
	fnext,
	Xfinfo,
	bfseek,
	XPR_EXTENSION,
	NULL,			/* Malloced for us */
	XOptions,
	unlink,
	NULL,
	NULL
    };

    int n, d0 = 0;
    unsigned char c, done;
    char xPrefs[256];
    char PrefFile[64], *p;
    int fchid;
    short msgno;
    char *pdev;
    
    /*
     * Set pipes unbuffered
     */
     
    setnbf (stdin);
    setnbf (stdout);

    GetXPRThing();

    if(!pXB)
    {
        BombOut();
    }
    strcpy (sx.xpru_protocol, pXB->xprname);
    pdev = pXB->libdir;

    /*
     * Try and open the preference files
     */
     
    fchid = GetPrefsFile(PrefFile, sx.xpru_protocol, pdev, 0);
    
    if (fchid)
    {
        int n;
	if (io_fline(fchid, -1, Prefs, (short) sizeof (Prefs)) > 0)
	{
	    if ((p = strchr (Prefs, '\n')))
	    {
		*p = '\0';
	    }
	    io_close (fchid);
	    X.xpr_filename = Prefs;
	    d0 = XProtocolSetup (&X);
	}
    }
    else
    {
	X.xpr_filename = NULL;
	X.xpr_options = NULL;
	X.xpr_gets = Xgets;
	d0 = XProtocolSetup (&X);
	X.xpr_options = XOptions;
	X.xpr_gets = MyGets;
    }

    OPipe = getchid (1);
    IPipe = getchid (0);

    n = io_fstrg (IPipe, -1, (char *) &msgno, 2);

    /*
     * This should catch applications using XPR v1.x
     * (or < 2.n, n < 3) calls
     */
     
    if (n != 2 || msgno != XMSG_START)
    {
        BombOut();
    }

    pS = pXB->psaddr;

    /*
     * d0 contains our status (& capabilities)
     * so we return it.
     */
     
    pXB->status = d0;
    io_sstrg (OPipe, -1, (char *) &msgno, 2);

    /*
     * Now loiter waiting for a client
     */
    
    for (done = 0; !done && (io_fstrg (IPipe, -1, (char *) &msgno, 2) == 2);)
    {
        short iwait, iflow;
        char icis;

        /*
         * The initial status gives us a clue as to the I/O to use
         */
         
        IOFlag = pXB->status;
        
	switch (msgno)
	{
	    case XMSG_SETUP:
		X.xpr_filename = NULL;
		X.xpr_options = XOptions;
		strcpy (xPrefs, Prefs);
		d0 = XProtocolSetup (&X);
		X.xpr_options = NULL;
		X.xpr_gets = Xgets;
		XProtocolSetup (&X);
		X.xpr_gets = MyGets;
		if (stricmp (Prefs, xPrefs))
		{
		    if(*PrefFile)
		    {
		        fchid = io_open(PrefFile, 3);
		    }
		    else
		    {
                        fchid = GetPrefsFile(PrefFile, sx.xpru_protocol, pdev, 3);
                    }
		    if (fchid > 0)
		    {
			io_sstrg (fchid, -1, Prefs, (short) strlen(Prefs));
			io_sbyte (fchid, -1, '\n');
			io_close (fchid);
		    }
		    else
		    {
		    /* Show an error */
		    }
		}
		break;
	    case XMSG_SEND:
		X.xpr_filename = GetXParams ();
		sx_dirn = 0;
		iflow = pS->bFlow;
		iwait = pS->Wait;
                icis = pS->QCIS;
		pS->Wait = 10;
		d0 = XProtocolSend (&X);
		LogInfo ();
		pS->bFlow = iflow;
		pS->Wait = iwait;
                pS->QCIS = icis;
		break;
	    case XMSG_RECV:
		X.xpr_filename = GetXParams ();
		sx_dirn = 1;
                icis = pS->QCIS;
		iflow = pS->bFlow;
		d0 = XProtocolReceive (&X);
		LogInfo ();
		pS->bFlow = iflow;
                pS->QCIS = icis;
		break;
	    case XMSG_DONE:
                done = 1;
		d0 = XProtocolCleanup (&X);
		break;
	}

        /*
         * Return the result
         */

	if(!done)
	{
	    /*
	     * If the client is a PTR program, it will have set up
	     * pXB->chid and pXB->job. We pick the chan/job and move the
             * pointer, just to make sure it's paying attention.
             */
         
	    if (pXB->chid > 0 && pXB->job > 0)
	    {
                
	        while (pXB->icon == 0)
	        {
		    mt_susjb (-1, 2, NULL);
	        }
	        pXB->check = 0x10000;
	        DoWake ();
	        while (pXB->flag == 1)
	        {
		    mt_susjb (-1, 1, NULL);
	        }
	        if (pXB->flag == 2)
	        {
		    pXB->flag = 0;
		    Xupdate (&sx);
		    DoWake ();
	        }
	    }
	}
	msgno = XMSG_STATUS;
	pXB->status = d0;
	io_sstrg (OPipe, -1, (char *) &msgno, 2);
	pXB->job = pXB->chid = pXB->buf = 0;
	sptr = 0;
    }
}
DoWake ()
{
    pick (pXB->chid, -1, pXB->job, 8L);
    if (sptr)
    {
	short x = 220, y = 60;
	setptr (pXB->chid, -1, &x, &y, -1);
    }
}

/*
 * Writes log info to the log file
 */
 
LogInfo ()
{
    char txt[256];

    if (pXB->logid > 0 && *sx.xpru_filename)
    {
        int bytes, rate;
        if(sx.xpru_bytes)
        {
            bytes = (sx.xpru_filesize) ? sx.xpru_filesize : sx.xpru_bytes;
            rate = sx.xpru_datarate;
        }
        else
        {
            bytes = rate = 0;
        }
	sprintf (txt, "%s %s %s\n", sx.xpru_protocol,
		 (sx_dirn) ? "Download" : "Upload",
		 sx.xpru_filename);
	io_sstrg (pXB->logid, -1, txt, (short) strlen (txt));
	sprintf (txt, "%d bytes %d char/sec %s\n\n",
		 bytes, rate,
		 sx.xpru_msg);
	io_sstrg (pXB->logid, -1, txt, (short) strlen (txt));
    }
}

/*
 * Copies a specified number of characters
 */

static char *maxcpy (char *d, char *s, int n)
{
    char c;
    char *base = d;

    while (n && (c = *s))
    {
	*d = c;
	++s;
	++d;
	--n;
    }
    *d = 0;
    return (base);
}
