/** xprascii.c
*
*   These are the protocol transfer routines for a simple ASCII upload.*
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/*
*   xproto.h is the include file given in Appendix B.
*/
#include "xproto.h"
#include "xprascii.h"

/**
*   Send a file
**/

UpdateRate (PXIO IO, struct XPR_UPDATE * u)
{
    int secs;
    static char buff1[24], buff2[24];
    static char *timefmt = "%02ld:%02ld:%02ld";

    time (&secs);

    secs -= XD->stv;

    if (XD->stv && secs > 0)
    {
	u->xpru_updatemask |= XPRU_ELAPSEDTIME;
	sprintf (buff1, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);
	u->xpru_elapsedtime = buff1;
	if (XD->ffc)
	{
	    u->xpru_datarate = XD->ffc / secs;
	    u->xpru_updatemask |= XPRU_DATARATE;
	}
	else
	{
	    u->xpru_datarate = 0;
	}
    }
    /*
     * If we know the file size and the data rate we can compute the
     * estimated time.
     */

    if (XD->xfs)
    {
	if (u->xpru_datarate)
	{
	    u->xpru_updatemask |= XPRU_EXPECTTIME;
	    secs = XD->xfs / u->xpru_datarate;
	    sprintf (buff2, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);
	    u->xpru_expecttime = buff2;
	}
    }
}

char *FBuffer(f) int f;
{
    int n;
    char *v;
    
    for(n = 16384; n > 1024; n = n >> 2)
    {
        if((v = malloc((unsigned)n)))
        {
            setvbuf((FILE *)f, v, _IOFBF, (unsigned) n);
            break;
        }
    }
    return v;
}

long XProtocolSend(IO)
struct XPR_IO *IO;
{
   long r, i;
   long brkflag = 0, fl = 0L, sd = 0L;
   long (*xupdate)(), (*xswrite)(), (*xfopen)(), (*xfclose)(), (*xfread)(),
        (*xsread)(),  (*xchkabort)(), (*xfinfo)();
   unsigned *serbuff = NULL;
   struct XPR_UPDATE xpru;
    char *p;
/*
*   These are the call-backs we need. If any of them isn't provided, quit.
*   Could do some error reporting if at least xupdate is there.
*/
   if ((xupdate    = IO->xpr_update)   == NULL) return(0L);
   if ((xswrite    = IO->xpr_swrite)   == NULL) return(0L);
   if ((xfopen     = IO->xpr_fopen)    == NULL) return(0L);
   if ((xfclose    = IO->xpr_fclose)   == NULL) return(0L);
   if ((xfread     = IO->xpr_fread)    == NULL) return(0L);
   if ((xsread     = IO->xpr_sread)    == NULL) return(0L);
   if ((xchkabort  = IO->xpr_chkabort) == NULL) return(0L);
   if ((xfinfo     = IO->xpr_finfo) == NULL) return(0L);   
/*
*   Allocate a few buffers.
*/
   ASetup(IO);
   XD->buffer =  malloc((unsigned) 80L);
   serbuff = malloc((unsigned) 80L);
/*
*   If we ran out of memory, print a message.
*/
   if (XD->buffer == NULL || serbuff == NULL) {
      xpru.xpru_updatemask = XPRU_ERRORMSG;
      xpru.xpru_errormsg   = "Ran out of memory!";
      (*xupdate)( &xpru);
      return(0L);
   }
/*
*   Read the send delay, if a XProtocolSetup() was done before.
*   If send delay is too large, cut it off at 10 seconds.
*   In this example, the xpr_data field contains a null terminated string
*   containing the number of ticks to delay each 80 characters.
*/

/*
*   Open the file. One could do wild card detection here.
*/
   XD->xfs = (*xfinfo)(IO->xpr_filename, 1);
   XD->fp = (*xfopen)( IO->xpr_filename, "r");
   if (!XD->fp) {
      free(XD->buffer);
      free(serbuff);
      xpru.xpru_updatemask = XPRU_ERRORMSG | XPRU_FILENAME;
      xpru.xpru_errormsg   = "Failed to open input file";
      xpru.xpru_filename   = IO->xpr_filename;
      (*xupdate)( &xpru);
      return(0L);
   }
    XD->stv = time(NULL);
    XD->blklen = 80;
    XD->Want = 0;
    XD->Last = 0;
/*
*   Start the transfer. See 3.8 for a discussion on how to implement
*   xupdate. 
*/
   xpru.xpru_updatemask = XPRU_MSG | XPRU_FILENAME | XPRU_FILESIZE |
                            XPRU_PACKETDELAY | XPRU_CHARDELAY; 

   xpru.xpru_msg        = "Starting ASCII Send";
   xpru.xpru_chardelay  = XD->CDelay;
   xpru.xpru_packetdelay= XD->LDelay;   
   xpru.xpru_filename   = IO->xpr_filename;
   xpru.xpru_filesize   = XD->xfs;
   (*xupdate)( &xpru);
/*
*   Now read 80 byte chunks from the file using xfread.
*   xfread requires four arguments, a0, d0, d1 and a1.
*/
   xpru.xpru_blocks = 0L;

   while (r = filbuf(IO))
   {
        for (i = 0, p = XD->buffer; i < r; i++, p++)
        {
            (*xswrite)(p, 1);
	    if (*p == '\n')
            {
                if(XD->LDelay)
	        {
	            mt_susjb(-1, XD->LDelay, NULL);
	        }	
	    }
	    else
	    {

	        if(XD->CDelay)
	        {
	            mt_susjb(-1, XD->CDelay, NULL);
	        }
	    }
	
        }
        fl += r;
        xpru.xpru_updatemask = XPRU_BYTES | XPRU_BLOCKS | XPRU_BLOCKSIZE;
        xpru.xpru_bytes      = fl;
        xpru.xpru_blocks++;
        xpru.xpru_blocksize  = r;
        XD->ffc = fl;
        UpdateRate(IO, &xpru);
        (*xupdate)( &xpru);

/*
*   Eat any characters that might arrive from the serial port.
*   We're not really waiting for any characters: use a timeout of 0L.
*/
        while ((*xsread)( serbuff, 80L, 0L) > 0L) ;
/*
*   Check for "abort" here. Perhaps should call chkmisc() as well.
*/
        if (brkflag = xchkabort()) break;
    }
/*
*   Close the file
*/
   (*xfclose)(XD->fp);
   free(XD->buffer);
   free(serbuff);
/*
*   If we got here through chkabort() say Aborted.
*/
   xpru.xpru_updatemask       = XPRU_MSG;
   if (brkflag) xpru.xpru_msg = "Aborted";
   else         xpru.xpru_msg = "Done"; 
   (*xupdate)( &xpru);
   if (brkflag) return(0L);
   else		return(1L);
}

/**
*
*   Receive a file.
*
**/
long XProtocolReceive(IO)
struct XPR_IO *IO;
{
   long r, i;
   long brkflag = 0, fl = 0L, sd = 0L;
   long (*xupdate)(), (*xswrite)(), (*xfopen)(), (*xfclose)(), (*xfwrite)(),
        (*xsread)(),  (*xchkabort)(), (*xfinfo)();
   unsigned char *serbuff = NULL;
   struct XPR_UPDATE xpru;
    char c;
/*
*   These are the call-backs we need. If any of them isn't provided, quit.
*   Could do some error reporting if at least xupdate is there.
*/
   if ((xupdate    = IO->xpr_update)   == NULL) return(0L);
   if ((xswrite    = IO->xpr_swrite)   == NULL) return(0L);
   if ((xfopen     = IO->xpr_fopen)    == NULL) return(0L);
   if ((xfclose    = IO->xpr_fclose)   == NULL) return(0L);
   if ((xfwrite    = IO->xpr_fwrite)   == NULL) return(0L);
   if ((xsread     = IO->xpr_sread)    == NULL) return(0L);
   if ((xchkabort  = IO->xpr_chkabort) == NULL) return(0L);
   if ((xfinfo     = IO->xpr_finfo)    == NULL) return(0L);
/*
*   Allocate a buffer.
*/
   ASetup(IO);
   XD->buffer = malloc((unsigned) 80L);
/*
*   If we ran out of memory, print a message.
*   The argument needs to go in A0: calla does this for us. 
*/
   if (XD->buffer == NULL) {
      xpru.xpru_updatemask = XPRU_ERRORMSG;
      xpru.xpru_errormsg   = "Ran out of memory!";
      (*xupdate)( &xpru);
      return(0L);
   }
/*
*   Open the file. One could do wild card detection here.
*   xfopen requires two arguments, in a0 and a1 respectively.
*   Again, this must be done in assembler, and callaa does it.
*/
   XD->fp = (*xfopen)( IO->xpr_filename, "w");
   if (!XD->fp) {
      free(XD->buffer);
      xpru.xpru_updatemask = XPRU_ERRORMSG | XPRU_FILENAME;
      xpru.xpru_errormsg   = "Failed to open output file";
      xpru.xpru_filename   = IO->xpr_filename;
      (*xupdate)( &xpru);
      return(0L);
   }
/*
*   Start the transfer. See 3.8 for a discussion on how to implement
*   xupdate. 
*/
   XD->stv = time(NULL);
   XD->xfs = 0;
   xpru.xpru_updatemask = XPRU_MSG | XPRU_FILENAME;
   xpru.xpru_msg        = "Starting ASCII Receive";
   xpru.xpru_filename   = IO->xpr_filename;
   (*xupdate)( &xpru);
/*
*   Now read 80 byte chunks from the serial port using xsread. Stop
*   when no characters arrive for 5 sec.
*/
   xpru.xpru_blocks = 0L;
   while ((XD->blklen = (*xsread)(XD->buffer, 80L, 50L)) > 0L)
   {    
      r = putsec(IO);
      fl += XD->blklen;
      XD->ffc = fl;
      xpru.xpru_updatemask = XPRU_BYTES | XPRU_BLOCKS | XPRU_BLOCKSIZE;
      xpru.xpru_bytes      = fl;
      xpru.xpru_blocks++;
      xpru.xpru_blocksize  = XD->blklen;

      UpdateRate(IO, &xpru);
      (*xupdate)( &xpru);
      if (brkflag = xchkabort()) break;
   }
/*
*   Close the file
*/
   (*xfclose)(XD->fp);
   free(XD->buffer);
/*
*   If we got here through chkabort() say Aborted.
*/
   xpru.xpru_updatemask       = XPRU_MSG;
   if (brkflag) xpru.xpru_msg = "Aborted";
   else         xpru.xpru_msg = "Done"; 
   (*xupdate)( &xpru);
   if (brkflag) return(0L);
   else		return(1L);
}

/**
*
*   Cleanup
*
**/
long XProtocolCleanup(PXIO IO)
{
   if (IO->xpr_data) free(IO->xpr_data);
   IO->xpr_data = NULL;

   return(1L);
}

filbuf (PXIO IO)
{
    char c;
    int m;
    char *p = XD->buffer;
    
    if (!XD->EOL && !XD->Pad)
    {
	m = (*IO->xpr_fread) (p, 1, XD->blklen, XD->fp);
	if (m <= 0)
	    return 0;
	return m;
    }
    m = 0;
    if ((XD->Want & 2) && XD->EOL)
    {
	*p++ = '\r';
	m++;
    }
    if (XD->Want)
    {
	*p++ = '\n';
	m++;
    }
    XD->Want = 0;
    while ((*IO->xpr_fread) (&c, 1, 1, XD->fp) == 1)
    {
	if (c == '\n')
	{
	    if(XD->Last == '\n' && XD->Pad)
	    {
	        *p++ = ' ';
	        if (++m == XD->blklen)
	        {
		    XD->Want = 3;
		    break;
	        }
	    }
	    if(XD->EOL)
	    {
                *p++ = '\r';
	        if (++m == XD->blklen)
	        {
	            XD->Last = c;
		    XD->Want = 1;
		    break;
	        }
	    }
	}
	*p++ = c;
	XD->Last = c;
        if (++m == XD->blklen)
	{
	    break;
	}
    }
    return m;
}
putsec (IO)
PXIO IO;
{

    char *p;
    int n;
    int m;
    
    if ((n = XD->blklen) == 0)
	return 0;
    if (!XD->EOL)
    {
	m = (*IO->xpr_fwrite) (XD->buffer, 1, n, XD->fp);
    }
    else
    {
	for (m = 0, p = XD->buffer; --n >= 0; ++p)
	{
	    if (*p == '\r')
		continue;
	    (*IO->xpr_fwrite) (p, 1, 1, XD->fp);
	    m++;
	}
    }
    return m;
}
