/** xprkermit.c
*
*	This code is copyright 1990 by Stephen Walton and Marco Papa.  It may
*	be freely distributed in its original or in modified form, provided
*	this copyright notice is kept intact.  It may not be sold for profit,
*	but may be included as part of a commercial program provided that
*	said inclusion does not increase the cost of that program beyond a
*	modest handling fee.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef QDOS

#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};
char _prog_name[] = "<QDOS XPR-KERMIT v2.30>";
static QVS  (_qcfx_, "QCFX>>01\x00\x09XPRKERMIT\x00\x00\x042.30\xFF\xFF");
#endif

#define XPRKERMIT 1
/*
 * xproto.h is the include file given in Appendix B.  It is included
 * in xprkermit.h
 */

#include "xprkermit.h"
#include "kermitproto.h"
#include "version.h"

long            (*xupdate) (), (*xswrite) (), (*xfopen) (), (*xfclose) (), (*xfread) (),
                (*xsread) (), (*xchkabort) (), (*xfnext) (), (*xffirst) (), (*xsflush) (),
                (*xfwrite) (), (*xgets) (), (*xfinfo) (), (*xunlink)() ,
		(*xsquery) (), (*xchkmisc) ();

extern char     start;		/* The start state for the protocol. */

/*
 * Forward declarations
 */
static SetupVars *setup(struct XPR_IO *IO);

/*
 * The flags for the kermitproto.w module.
 */
char           *p_pattern;	/* wildcard pattern */
int             parity;		/* parity on? 0 for no parity--need for
				 * proper 8th-bit quote */
int             text;		/* Text or binary mode? Flag 1 for text file,
				 * 0 for binary file	 */
int             convert;	/* Convert file names to lower case? 0 for
				 * literal files (no), 1 for translate (yes) */
int             urpsiz;		/* Kermit maximum packet size. Maximum
				 * receive packet size user wants.	 */
char           *cmarg;		/* Character string containing Kermit server
				 * cmd */

char            *defdir;

extern int      bctr;		/* Block check type to request. */
extern int      limit;		/* Retry limit.  May increase on very noisy
				 * lines. */
extern int      rtimo;		/* Timeout to request. */
extern int	keep;		/* Keep incomplete files? */

int             getfile;	/* Host server. 0 = Receive (no) ; 1 = Get
				 * (yes) */
long            brkflag;

static SetupVars Defaults = {
    {0},			/* No default file name. */
#ifdef QDOS
    {"ram2_"},
#else
    {"t:"},
#endif
    0,				/* Parity defaults to off. */
    1,				/* Text file defaults to on. */
    1,				/* Convert file names by default. */
    94,				/* Default maximum packet length. */
    1,				/* Default block check type. */
    5,				/* Retry limit */
    10,				/* Timeout (seconds) */
    0,				/* Don't keep incomplete files */
    0,				/* Host is server?  No by default. */
};

/*
 * Local prototypes.
 */
static KermitCd(struct XPR_IO *IO, char *dir);
static KermitFinish(struct XPR_IO *IO);
static KermitBye(struct XPR_IO *IO);
static DoGeneric(struct XPR_IO *IO, char *s);


/**
*
*   Send a file
*
**/
long
XProtocolSend(IO)
    struct XPR_IO  *IO;
{
    struct XPR_UPDATE xpru;
    SetupVars      *sv;

    if ((sv = setup(IO)) == NULL)
	return 0L;		/* Initialize parameters. */

    brkflag = 0;

    /*
     * Read the text/binary set using xpr_finfo if present.  Else use the
     * value chosen in XPRotocolSetup.
     */
    if (xfinfo) {
	/*
	 * Use feature that calling xpr_finfo with a zero-length filename
	 * returns setting of internal comm program Text/Binary flag.
	 */
	text = ((*xfinfo)(IO->xpr_filename, 2L) == 1 ? 0 : 1);
    }
    /*
     * Start the transfer. See 3.8 for a discussion on how to implement
     * xupdate.
     */
    tchar('#');

    /*
     * Copy filename, and put pointer in external Kermit variable.
     */
    strcpy(sv->FileName, IO->xpr_filename);
    p_pattern = sv->FileName;

    /*
     * */
    start = 's';
    proto();

    /*
     * If we got here through chkabort() say Aborted.
     */
    xpru.xpru_updatemask = XPRU_MSG;
    if (brkflag)
	xpru.xpru_msg = "Aborted";
    else
	xpru.xpru_msg = "Done";
    (void) (*xupdate) (&xpru);
    if (brkflag)
	return (0L);
    else
	return (1L);
}


/**
*
*   Receive a file.
*
**/
long
XProtocolReceive(IO)
    struct XPR_IO  *IO;
{
    struct XPR_UPDATE xpru;
    long            status;
    SetupVars      *sv;

    if ((sv = setup(IO)) == NULL)
	return 0L;			/* Initialize parameters. */

    brkflag = 0;

    /*
     * Read the text/binary set using xpr_finfo if present.  Else use the
     * value chosen in XPRotocolSetup.
     */
    if (xfinfo) {
	/*
	 * Use feature th at calling xpr_finfo with a zero-length filename
	 * returns setting of internal comm program Text/Binary flag.
	 */
	text = ((*xfinfo)("", 2L) == 1 ? 0 : 1);
    }
    /*
     * Start the transfer. See 3.8 for a discussion on how to implement
     * xupdate.
     */
    tchar('#');

    if (getfile) {
	/*
	 * Copy filename, and put pointer in external Kermit variable.
	 */
	cmarg = sv->FileName;
	start = 'r';
	status = (*xgets) ("Host Filename", cmarg);
	if (!status)
	    return (0L);
    } else
	start = 'v';


    defdir  = sv->defdir;
    
    proto();

    /*
     * If we got here through chkabort() say Aborted.
     */
    xpru.xpru_updatemask = XPRU_MSG;
    if (brkflag)
	xpru.xpru_msg = "Aborted";
    else
	xpru.xpru_msg = "Done";
    (void) (*xupdate)( &xpru);
    if (brkflag)
	return (0L);
    else
	return (1L);
}

/*
 * Perform a generic Kermit server command.
 */
static
DoGeneric(IO, s)
    struct XPR_IO  *IO;
    char           *s;
{
    if (setup(IO) == NULL)
	return 0;			/* Set up transfer characteristics. */
    brkflag = 0;
    start = 'g';
    cmarg = s;
    proto();
    if (brkflag)
	return (0);
    else
	return (1);
}

/*
 * Execute a Kermit FINISH command.
 */
static
KermitFinish(IO)
    struct XPR_IO  *IO;
{
    return (DoGeneric(IO, "F"));
}

/*
 * Execute a Kermit BYE command.
 */
static
KermitBye(IO)
    struct XPR_IO  *IO;
{
    return (DoGeneric(IO, "L"));
}

/*
 * Change directory on remote server.  We need to return an error if the CD
 * fails on the remote end.
 */
static
KermitCd(IO, dir)
    struct XPR_IO  *IO;
    char           *dir;
{
    char            CdCommand[100];
    int             retval;

    CdCommand[0] = 'C';
    CdCommand[1] = (char) tochar(strlen(dir));
    strcpy(CdCommand + 2, dir);
    retval = DoGeneric(IO, CdCommand);
    return retval;
}

/**
*
*   Setup
*
* First, a general-purpose comparison for either of the two possible returns
* indicating a Yes push on a Boolean gadget.
**/

#define XprBoolTrue(s) ((*s == 'Y') || (*s == 'y') || (stricmp(s, "on") == 0))

/*
 * Then, a small set of code to initialize a string based on a value.
 */


static char     YesString[] = "yes";
static char     NoString[] = "no";

#define XprSet(value, string) if(value) strcpy(string, YesString); else strcpy(string, NoString)

/*
 * If Setup() succeeds, flag same.  Also, we don't need a file requester
 * on receive.
 */
#define SUCCESS (XPRS_SUCCESS | XPRS_NORECREQ)

static SetupVars *
InitData(struct XPR_IO *IO) {
    SetupVars *sv;

    /*
     * Allocate memory for file name buffer and options if first call.  Also
     * allocate user's data segment and point us at it.
     */
    if ((sv = IO->xpr_data) == NULL) {
	if ((sv = (SetupVars *) malloc(sizeof(SetupVars))) == NULL)
	    return NULL;
	*sv = Defaults;
	IO->xpr_data = sv;
    }
    return(sv);
}

long
XProtocolSetup(IO)
    struct XPR_IO  *IO;
{
    long            (*xoptions) ();
#define NOPTS 10
    struct xpr_option opt[NOPTS];
#define MAXSTRING 6
    char            ValueStrings[NOPTS][MAXSTRING];
    long            status;
    int             i, j;
    char            buf[256];
    SetupVars      *sv, tempvar;

    if ((sv = InitData(IO)) == NULL)
        return (XPRS_FAILURE);

    if ((xupdate = IO->xpr_update) == NULL)
	return (XPRS_FAILURE);
    if ((xgets = IO->xpr_gets) == NULL)
	return (XPRS_FAILURE);

    tempvar = *sv;
    /*
     * In order to use xpr_options, we must have all of the three conditions
     * which follow true.  Otherwise, either IO->xpr_filename is non-NULL, in
     * which case we assume it contains a setup string, or there is no
     * xpr_options, in which case we use xpr_gets to retrieve a setup string.
     */
    if (IO->xpr_filename == NULL &&
	IO->xpr_extension >= 1 &&
	(xoptions = IO->xpr_options) != NULL) {

	/*
	 * I use a counter, i, here, so that I can stick in more options
	 * without needing to change a lot of numbers.  I can also skip
	 * options dynamically, as in Text below.
	 */
	i = 0;
	/* Announce us. */
	opt[i].xpro_description = "Kermit Commands " VSTRING;
	opt[i].xpro_type = XPRO_HEADER;
	opt[i].xpro_value = NULL;
	opt[i].xpro_length = 0;
	i++;
	/*
	 * First, do the Kermit server command items.
	 */
	opt[i].xpro_description = "Kermit FINISH";
	opt[i].xpro_type = XPRO_COMMAND;
	opt[i].xpro_value = NULL;
	opt[i].xpro_length = 0;
	i++;
	opt[i].xpro_description = "Kermit BYE";
	opt[i].xpro_type = XPRO_COMMAND;
	opt[i].xpro_value = NULL;
	opt[i].xpro_length = 0;
	i++;
	opt[i].xpro_description = "Kermit CD";
	opt[i].xpro_type = XPRO_COMMPAR;
	buf[0] = '\0';
	opt[i].xpro_value = buf;		/* Use buf to hold dir name. */
	opt[i].xpro_length = sizeof(buf) - 1;
	i++;
	opt[i].xpro_description = "Kermit Options";
	opt[i].xpro_type = XPRO_COMMAND;
	opt[i].xpro_value = NULL;
	opt[i].xpro_length = 0;
	i++;
	/* show requester after loading pointers */

	status = (*xoptions) (opt, i);

	/* check returned value */
	if (status == -1L)
	    return XPRS_FAILURE;
	/* Check returned value to see what we are to do. */
	i = 1;
	if (status & (1L << i))
	    return (KermitFinish(IO) ? SUCCESS : XPRS_FAILURE);
	i++;
	if (status & (1L << i))
	    return (KermitBye(IO) ? SUCCESS : XPRS_FAILURE);
	i++;
	if (status & (1L << i))
	    return (KermitCd(IO, opt[i].xpro_value) ? SUCCESS : XPRS_FAILURE);
	/* If we get to this point, we are to set options. */
	i = 0;			/* Start over. */
	opt[i].xpro_description = "Kermit Options " VSTRING;
	opt[i].xpro_type = XPRO_HEADER;
	opt[i].xpro_value = NULL;
	opt[i].xpro_length = 0;
	i++;
	/* Convert filename */
	opt[i].xpro_description = "Convert Filename";
	opt[i].xpro_type = XPRO_BOOLEAN;
	XprSet(tempvar.ConvertFlag, ValueStrings[i]);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/* Partial file keep */
	opt[i].xpro_description = "Keep Incomplete";
	opt[i].xpro_type = XPRO_BOOLEAN;
	XprSet(tempvar.KeepFlag, ValueStrings[i]);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/* host is server */
	opt[i].xpro_description = "Host Server";
	opt[i].xpro_type = XPRO_BOOLEAN;
	XprSet(tempvar.GetFlag, ValueStrings[i]);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/*
	 * file type -- only show this if xpr_finfo not present, or returns
	 * error when we try to get the file type.
	 */
	if ((xfinfo = IO->xpr_finfo) == NULL ||
	    (tempvar.TextFlag = (int) ((*xfinfo)("", 2L))) == 0) {
	    tempvar.TextFlag = -1;	/* Flag xfinfo failed. */
	    opt[i].xpro_description = "Text File";
	    opt[i].xpro_type = XPRO_BOOLEAN;
	    XprSet(tempvar.TextFlag, ValueStrings[i]);
	    opt[i].xpro_value = ValueStrings[i];
	    opt[i].xpro_length = MAXSTRING;
	    i++;
	} else
	    /*
	     * Switch to 0 for binary, 1 for text; xpr_finfo returns 1 for
	     * binary, 2 for text.
	     */
	    tempvar.TextFlag -= 1;
	/* Packet size */
	opt[i].xpro_description = "Packet Size";
	opt[i].xpro_type = XPRO_LONG;
	(void) sprintf(ValueStrings[i], "%-d", tempvar.MaxPacket);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/* Block Check type */
	opt[i].xpro_description = "Block Check (1, 2, 3)";
	opt[i].xpro_type = XPRO_LONG;
	(void) sprintf(ValueStrings[i], "%-d", tempvar.BlockCheckType);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/* Retry Limit */
	opt[i].xpro_description = "Maximum Retries";
	opt[i].xpro_type = XPRO_LONG;
	(void) sprintf(ValueStrings[i], "%-d", tempvar.RetryLimit);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	/* Timeout */
	opt[i].xpro_description = "Timeout (seconds)";
	opt[i].xpro_type = XPRO_LONG;
	(void) sprintf(ValueStrings[i], "%-d", tempvar.Timeout);
	opt[i].xpro_value = ValueStrings[i];
	opt[i].xpro_length = MAXSTRING;
	i++;
	opt[i].xpro_description = "Download Dir";
	opt[i].xpro_type = XPRO_STRING;
	opt[i].xpro_value = tempvar.defdir;
	opt[i].xpro_length = sizeof(tempvar.defdir);
	i++;

	/* show requester */
	status = (*xoptions)(opt, i);

	/* check returned value */
	if (status == -1L)
	    return XPRS_FAILURE;
	i = 1;			/* Skip header		 */
	    tempvar.ConvertFlag = XprBoolTrue(opt[i].xpro_value) ? 1 : 0;
	i++;
	    tempvar.KeepFlag = XprBoolTrue(opt[i].xpro_value) ? 1 : 0;
	i++;
	    tempvar.GetFlag = XprBoolTrue(opt[i].xpro_value) ? 1 : 0;
	i++;
	if (xfinfo == NULL || tempvar.TextFlag == -1) {
		tempvar.TextFlag = XprBoolTrue(opt[i].xpro_value) ? 1 : 0;
	    i++;
	}
	    tempvar.MaxPacket = atoi(opt[i].xpro_value);
	i++;
	    tempvar.BlockCheckType = atoi(opt[i].xpro_value);
	i++;
	    tempvar.RetryLimit = atoi(opt[i].xpro_value);
	i++;
	    tempvar.Timeout = atoi(opt[i].xpro_value);
    } else {
	if (IO->xpr_filename != NULL)
	    strcpy(buf, IO->xpr_filename);	/* Save setup string */
	else {			/* Prompt for command/options string */
	    sprintf(buf, "OC%c,G%c,K%c,T%c,P%d,B%d,R%d,O%d,D%s",
                    tempvar.ConvertFlag ? 'Y' : 'N',
		    tempvar.GetFlag ? 'Y' : 'N', tempvar.KeepFlag ? 'Y' : 'N',
                    tempvar.TextFlag ? 'Y' : 'N', tempvar.MaxPacket,
	       tempvar.BlockCheckType, tempvar.RetryLimit, tempvar.Timeout,
                tempvar.defdir);
	    if ((*xgets)("Kermit Options", buf) == 0)
		return XPRS_FAILURE;	/* Failed to set up? */
	}
	(void) strupr(buf);
	switch (buf[0]) {
	case 'F':
	    return (KermitFinish(IO) ? SUCCESS : XPRS_FAILURE);
	case 'B':
	    return (KermitBye(IO) ? SUCCESS : XPRS_FAILURE);
	case 'C':
	    return (KermitCd(IO, buf + 1) ? SUCCESS : XPRS_FAILURE);
	case 'O':
	    if (SetupFromString(IO, buf, &tempvar) == 0)
		return XPRS_FAILURE;
	    break;
	case '\0':
	    break;
	default:
	    ioerr(IO, "Unrecognized XPR Kermit setup string");
	    break;
	}
    }
    *sv = tempvar;		/* Copy setups into safe place. */
    /*
     * Return success and inform caller that we don't need a requester for
     * receive.
     */
    return SUCCESS | XPRS_NORECREQ;
}

static char Delimiters[] = " \t\r\n,";

int
SetupFromString(IO, s, sv)
    struct XPR_IO  *IO;
    char           *s;
    SetupVars      *sv;
{
    char           *p;
    char            errbuf[50];

    if (*s != 'O')
	return 0;		/* Options string must start with O. */
    s++;			/* Skip leading O. */
    /*
     * Hunt for options with strtok.  We allow whitespace and commas to
     * separate options.
     */
    for (p = strtok(s, Delimiters); p != NULL; p = strtok(NULL, Delimiters)) {
	switch (*p++) {				/* Auto-increment to option */
	case 'C':				/* Case conversion. */
	    if (*p == 'Y' || *p == 'N')
		sv->ConvertFlag = (*p == 'Y');
	    else
		ioerr(IO, "Illegal C option format (must be Y or N)");
	    break;
	case 'G':				/* Get files (Host server) */
	    if (*p == 'Y' || *p == 'N')
		sv->GetFlag = (*p == 'Y');
	    else
		ioerr(IO, "Illegal G option format (must be Y or N)");
	    break;
	case 'K':				/* Keep incomplete file */
	    if (*p == 'Y' || *p == 'N')
		sv->KeepFlag = (*p == 'Y');
	    else
		ioerr(IO, "Illegal K option format (must be Y or N)");
	    break;
	case 'T':				/* Text file */
	    if (*p == 'Y' || *p == 'N')
		sv->TextFlag = (*p == 'Y');
	    else
		ioerr(IO, "Illegal T option format (must be Y or N)");
	    break;
	case 'P':				/* Maximum packet length */
	    sv->MaxPacket = atoi(p);
	    break;
	case 'B':				/* Block check type. */
	    sv->BlockCheckType = atoi(p);
	    break;
	case 'R':				/* Retry limit */
	    sv->RetryLimit = atoi(p);
	    break;
	case 'O':				/* Timeout */
	    sv->Timeout = atoi(p);
	    break;
	case 'D':
            strcpy(sv->defdir, p);
            break;				/* Timeout */
	default:
	    sprintf(errbuf, "Illegal XPR Kermit option: %c", p[-1]);
	    ioerr(IO, errbuf);
	    break;
	}
    }
    return 1;
}

/**
*
*   Cleanup
*
**/
long
XProtocolCleanup(IO)
    struct XPR_IO  *IO;
{
    if (IO->xpr_data)
	free(IO->xpr_data);
    IO->xpr_data = NULL;

    return (1L);
}

int
XPRParity(IO)
    struct XPR_IO  *IO;
{
	return 0;		/* Assume no parity if can't tell. */
}

void
XPRLong(IO, i)
    struct XPR_IO  *IO;
    long            i;
{
    struct XPR_UPDATE xpru;
    char            locbuf[80];

    if ((xupdate = IO->xpr_update) == NULL)
	return;
    /* debug: show long value */
    xpru.xpru_updatemask = XPRU_MSG;
    sprintf(locbuf, "%lx", i);
    xpru.xpru_msg = &locbuf[0];
    (void) (*xupdate)(&xpru);
}

/*
 * Copy setup variables into the local copies.  If there are none, then
 * simply copy the defaults.
 */
static SetupVars * setup(IO)
    struct XPR_IO  *IO;
{
    register SetupVars *Current;

    if ((Current = InitData(IO)) == NULL)
        return NULL;
    /*
     * These are the call-backs we need. If any of them isn't provided, quit.
     * Could do some error reporting if at least xupdate is there.
     */
    if ((xupdate = IO->xpr_update) == NULL)
	return (0L);
    if ((xswrite = IO->xpr_swrite) == NULL)
	return (0L);
    if ((xfopen = IO->xpr_fopen) == NULL)
	return (0L);
    if ((xfclose = IO->xpr_fclose) == NULL)
	return (0L);
    if ((xfread = IO->xpr_fread) == NULL)
	return (0L);
    if ((xsread = IO->xpr_sread) == NULL)
	return (0L);
    if ((xchkabort = IO->xpr_chkabort) == NULL)
	return (0L);
    if ((xfnext = IO->xpr_fnext) == NULL)
	return (0L);
    if ((xffirst = IO->xpr_ffirst) == NULL)
	return (0L);
    if ((xsflush = IO->xpr_sflush) == NULL)
	return (0L);
    if ((xfwrite = IO->xpr_fwrite) == NULL)
	return (0L);
    if ((xgets = IO->xpr_gets) == NULL)
	return (0L);
    /*
     * Here are callbacks we can do without.
     */
    xchkmisc = IO->xpr_chkmisc;
    if (IO->xpr_extension >= 2)
	xunlink = IO->xpr_unlink;
    else
	xunlink = NULL;
    if (IO->xpr_extension >= 3)
        xsquery = IO->xpr_squery;
    else
	xsquery = NULL;
    xfinfo = IO->xpr_finfo;

    parity = Current->ParityFlag = XPRParity(IO);
    text = Current->TextFlag;
    convert = Current->ConvertFlag;
    urpsiz = Current->MaxPacket;
    bctr = Current->BlockCheckType;
    limit = Current->RetryLimit;
    rtimo = Current->Timeout;
    getfile = Current->GetFlag;
    keep = Current->KeepFlag;
    if(IO->xpr_setserial)
    {
        (*IO->xpr_setserial)(3);
    }
    return Current;
}

/*
 * Have the comm program display an error message for us, using a temporary
 * XPR_UPDATE structure; used to display errors before Vars gets allocated
 */
void
ioerr(IO, msg)
    struct XPR_IO  *IO;
    char           *msg;
{
    struct XPR_UPDATE xpru;

    if ((xupdate = IO->xpr_update) != NULL) {
	xpru.xpru_updatemask = XPRU_ERRORMSG;
	xpru.xpru_errormsg = msg;
	(void) (*xupdate)(&xpru);
    }
}
