/*
 * */

/* #define DEBUGLOG 1 */

/* Return codes */
#define OK        0
#define ERROR   (-1)
#define TIMEOUT (-2)
#define RCDO    (-3)

/* Relevant control characters */
#define CR	('M' & 0x1F)	/* ^M */
#define DLE	('P' & 0x1F)	/* ^P */
#define XON	('Q' & 0x1F)	/* ^Q */
#define XOFF	('S' & 0x1F)	/* ^S */
#define CAN	('X' & 0x1F)	/* ^X */
#define CPMEOF	('Z' & 0x1F)	/* ^Z */

/* Misc. program constants */
#define LZMANAG		  0	/* Default ZMODEM file management mode */
#define LZTRANS		  0	/* Default ZMODEM file transport mode */
#define PATHLEN		256	/* What's the max legal path size? */
#define CONFIGLEN	 32	/* Max length of transfer options string */
#define KSIZE	       1024	/* Max allowable packet size */
#define MINBLOCK	 64	/* Min allowable packet size */
#define MAXGOODNEEDED  8192	/* Max # good bytes req'd to bump packet size */

/* Provision for future 7-bit ZMODEM; for now, there's no difference */
#define sendline xsendline

/*
 * Replacement for global variables normally used, in order to make code
 * fully reentrant; each invocation allocs their own Vars, and passes the
 * struct pointer down through the entire program so they're always
 * available. Pointer to this struct is usually able to be a register
 * variable, so access is no worse than any stack variable (all
 * register-relative).  Kinda kludgey, but the original ZModem code design
 * depended on lots of globals, and I didn't want to redesign the whole damn
 * thing.  Besides, it's more efficient than constantly pushing & popping
 * args all over the place.
 */

struct Vars
{
    struct XPR_IO io;		/* Copy of XProto IO struct passed by term
				 * prog. */
    struct XPR_UPDATE xpru;	/* Scratchpad xpr_update() control struct */
    int Starttime;		/* Time transfer started */
    UBYTE Rxhdr[4];		/* Received header */
    UBYTE Txhdr[4];		/* Transmitted header */
    UBYTE Msgbuf[128];		/* Scratchpad buffer for printing messages */
    UBYTE Filename[PATHLEN];	/* Name of the file being up/downloaded */
    UBYTE Pktbuf[KSIZE];	/* File data packet buffer */
#ifndef QDOS
    UBYTE Modembuf[32];	        /* Input buffer for data from modem */
#endif
    UBYTE Outbuf[KSIZE * 2 + 256];	/* Output buffer for data to modem */
#ifndef QDOS
    UBYTE *Modemchar;		/* Next char to get from Modembuf */
    UBYTE *Filebuf;		/* File I/O buffer address */
    UBYTE *Filebufptr;		/* Current position within Filebuf */
    long Filebufpos;		/* File offset of data in Filebuf */
    long Filebufmax;		/* Size of Filebuf */
    long Filebuflen;		/* Number of bytes currently stored in
				 * Filebuf */
    long Filebufcnt;		/* Number of bytes remaining/written in
				 * Filebuf */
#endif
    long File;			/* Handle of file being transferred */
    long Oldstatus;		/* Original terminal program's modem settings */
    long Baud;			/* BPS setting of modem */
    long Strtpos;		/* Starting byte position of transfer */
    long Fsize;			/* Size of file being transferred */
    long Rxbytes;		/* Number of bytes received so far */
    long Rxpos;			/* Received file position */
    long Txpos;			/* Transmitted file position */
    short Filcnt;		/* Number of files opened for transmission */
    short Errcnt;		/* Number of files unreadable */
    short Noroom;		/* Flags 'insufficient disk space' errors */
    short Rxbuflen;		/* Largest frame they're willing to xfer */
    short Tframlen;		/* Largest frame we're willing to xfer */
    short Rxtimeout;		/* Tenths of seconds to wait for something */
    short Tryzhdrtype;		/* Header type to send corresp to Last rx
				 * close */
    short Modemcount;		/* Number of bytes available in Modembuf */
    short Outbuflen;		/* Number of bytes currently stored in Outbuf */
    short Rxframeind;		/* ZBIN or ZHEX; type of frame received */
    short Txfcs32;		/* TRUE means send binary frame with 32 bit
				 * FCS */
    short Rxflags;		/* Temp register */
    short Wantfcs32;		/* want to send 32 bit FCS */
    short Crc32t;		/* Display flag indicates 32 bit CRC being
				 * sent */
    short Crc32;		/* Display flag indicates 32 bit CRC being
				 * recd */
    short Rxtype;		/* Type of header received */
    short Rxcount;		/* Count of data bytes received */
    short Znulls;		/* Number of nulls to send at begin of ZDATA
				 * hdr */
    short ErrorLimit;		/* How many sequential errors before abort */
    char Rxbinary;		/* Force binary mode download? */
    char Rxascii;		/* Force text mode download? */
    char Thisbinary;		/* Receive this file in binary mode? */
    char Lzconv;		/* Suggested binary/text mode for uploads */
    char Eofseen;		/* Text-mode EOF marker (^Z) rec'd on
				 * download? */
    UBYTE Zconv;		/* ZMODEM file conversion request */
    UBYTE Zmanag;		/* ZMODEM file management request */
    UBYTE Ztrans;		/* ZMODEM file transport request */
    UBYTE Lastsent;		/* Last text char written by putsec() */
    UBYTE Lastzsent;		/* Last char sent by zsendline() */
#ifndef QDOS
    UBYTE Fileflush;		/* Flush file I/O buffer before closing? */
#endif
    UBYTE Attn[ZATTNLEN + 1];	/* Attention string rx sends to tx on err */
};

/*
 * Option settings and other variables needed outside of
 * XProtocolSend/Receive; separated from rest of Vars so that huge Vars
 * struct doesn't have to be allocated except during transfers.  Pointer to
 * this struct kept in xpr_data.
 */
struct SetupVars
{
#ifndef QDOS
    UBYTE *matchptr, *bufpos;
    short buflen;
#endif
    UBYTE option_t[2], option_o[2], option_b[8], option_f[8], option_e[8], option_s[4];
    UBYTE option_r[4], option_a[4], option_d[4], option_k[4], option_p[32];
};

/* Function prototypes */

long XProtocolSend (struct XPR_IO *);
short getzrxinit (struct Vars *);
void sendbatch (struct Vars *);
short sendone (struct Vars *);
short sendname (struct Vars *);
short zsendfile (struct Vars *, short);
short zsendfdata (struct Vars *);
short getinsync (struct Vars *);
void saybibi (struct Vars *);

long XProtocolReceive (struct XPR_IO *);
short rcvbatch (struct Vars *);
short tryz (struct Vars *);
short rzfiles (struct Vars *);
short rzfile (struct Vars *);
short procheader (struct Vars *);
short putsec (struct Vars *);
void ackbibi (struct Vars *);

long XProtocolSetup (struct XPR_IO *);
long XProtocolCleanup (struct XPR_IO *);
long XProtocolHostMon (struct XPR_IO *, char *, long, long);
long XProtocolUserMon (struct XPR_IO *, char *, long, long);
struct Vars *setup (struct XPR_IO *);
UBYTE *find_option (UBYTE *, UBYTE);
void set_textmode (struct Vars *);
void canit (struct Vars *);
void zmputs (struct Vars *, UBYTE *);
void xsendline (struct Vars *, UBYTE);
void sendbuf (struct Vars *);
short readock (struct Vars *, short);
char char_avail (struct Vars *);
void update_rate (struct Vars *);
long bfopen (struct Vars *, UBYTE *);
void bfclose (struct Vars *);
void bfseek (struct Vars *, long);
long bfread (struct Vars *, UBYTE *, long);
long bfwrite (struct Vars *, UBYTE *, long);
void ioerr (struct XPR_IO *, char *);
void upderr (struct Vars *, char *);
void updmsg (struct Vars *, char *);
long getfree (void);
char exist (struct Vars *);
int mysprintf ();

#ifdef DEBUGLOG
void dlog (struct Vars *, UBYTE *);

#endif

void zsbhdr (struct Vars *, USHORT);
void zshhdr (struct Vars *, USHORT);
void zsdata (struct Vars *, short, USHORT);
short zrdata (struct Vars *, UBYTE *, short);
short zrdat32 (struct Vars *, UBYTE *, short);
short zgethdr (struct Vars *);
short zrbhdr (struct Vars *);
short zrbhdr32 (struct Vars *);
short zrhhdr (struct Vars *);
void zputhex (struct Vars *, UBYTE);
void zsendline (struct Vars *, UBYTE);
short zgethex (struct Vars *);
short zdlread (struct Vars *);
short noxrd7 (struct Vars *);
void stohdr (struct Vars *, long);
long rclhdr (struct Vars *);




/* End of XprZmodem.h source */
