
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "xproto.h"
#include "xprxmodem.h"

#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};

char _prog_name[] = "<QDOS XPR-XMODEM v2.30>";
static QVS (_qcfx_, "QCFX>>01\x00\x09XPRXMODEM\x00\x00\x042.30\xFF\xFF");

char *find_option(char *, char);

static struct xpr_option op[] =
{
    {"XModem options:", XPRO_HEADER, NULL, 0},
    {"Text mode (Y,N,C):", XPRO_STRING, NULL, 0},
    {"CRC mode (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"1K Blocks (Y,N):", XPRO_BOOLEAN, NULL, 0},
};

static char option_t[4], option_c[4], option_k[4];

CheckSetup (IO)
PXIO IO;
{
    if (IO->xpr_data == NULL)
    {
	if ((IO->xpr_data = (PXPD) malloc (sizeof (XPD))) == NULL)
	{
	    ioerr (IO, "Not enough memory");
	    return ERROR;
	}
	setmem (IO->xpr_data, sizeof (XPD), '\0');
	XD->Ascii_conf = 'N';
	XD->crc_conf = 'Y';
	XD->Kblocks_conf = 'Y';
    }
    return OK;
}

/**********************************************************
 **********************************************************/

long XProtocolSetup (IO)
PXIO IO;
{

    char buf[80], *p;
    int i, len;

    if (CheckSetup (IO) == ERROR)
    {
	return XPRS_FAILURE;
    }

    /* If options string passed by comm prog, use it; else prompt user */
    if (IO->xpr_filename)
    {
	strcpy (buf, IO->xpr_filename);
    }
    else
    {
	/* If xpr_options() implemented by comm program, use it */
	if (IO->xpr_extension >= 1 && IO->xpr_options)
	{
	    op[1].xpro_value = option_t;
	    op[1].xpro_length = sizeof (option_t);
	    *option_t = XD->Ascii_conf;
	    *(option_t + 1) = '\0';
	    op[2].xpro_value = option_c;
	    op[2].xpro_length = sizeof (option_c);
	    *option_c = XD->crc_conf;
	    *(option_c + 1) = '\0';
	    op[3].xpro_value = option_k;
	    op[3].xpro_length = sizeof (option_k);
	    *option_k = XD->Kblocks_conf;
	    *(option_k + 1) = '\0';

	    /* Convert Y/N used elsewhere into "yes"/"no" required by spec */
	    for (i = 2; i < 4; ++i)
	    {
		strcpy (op[i].xpro_value, (op[i].xpro_value[0] == 'Y') ? "yes" : "no");
	    }

	    (*IO->xpr_options) (op, 4);

	    /* Convert "yes"/"no" or "on"/"off" into Y/N */
             for (i = 2; i < 4; ++i)
             {
                  if(op[i].xpro_value[0] == 'Y' ||
                    op[i].xpro_value[0] == 'y' ||
                    stricmp (op[i].xpro_value, "yes") == 0 ||
                    stricmp (op[i].xpro_value, "on") == 0)
                    {
                        op[i].xpro_value[0] = 'Y';
                    }
                    else
                    {
                        op[i].xpro_value[0] = 'N';
                    }

             }
            sprintf (buf, "C%c,K%c,T%c", *option_c, *option_k, *option_t);

	}
	else
	{
	    /* Start buffer with current settings so user can see/edit them. */
            sprintf (buf, "C%c,K%c,T%c", XD->crc_conf, XD->Kblocks_conf, XD->Ascii_conf);
	    if (IO->xpr_gets)
		(*IO->xpr_gets) ("XModem options:", buf);
	}
    }
    /* Upshift options string for easier parsing */
    strupr (buf);

    /*
     * Merge new T(ext) option into current settings if given "TY" = Force
     * Text mode on, "TN" = Force Text mode off, "T?" = Use other end's text
     * mode suggestion (default to binary) "TC" = Ask Comm program for file
     * type
     */
    if (p = find_option (buf, 'T'))
    {
	if (*p == 'Y' || *p == 'N' || *p == 'C')
	    XD->Ascii_conf = *p;
	else
	    ioerr (IO, "Invalid T flag ignored; should be Y, N, or C");
    }

    if (p = find_option (buf, 'C'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->crc_conf = *p;
	    XD->crcmode = (XD->crc_conf == 'Y');
	}
	else
	    ioerr (IO, "Invalid C flag ignored; should be Y or N");
    }

    if (p = find_option (buf, 'K'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->Kblocks_conf = *p;
	    XD->Kblocks = (XD->Kblocks_conf == 'Y');
	}
	else
	    ioerr (IO, "Invalid K flag ignored; should be Y or N");
    }

    return XPRS_SUCCESS;
}


XSetup (IO)
PXIO IO;
{
    if (!IO->xpr_update)
    {
	return ERROR;
    }
    if (!IO->xpr_fopen || !IO->xpr_fclose || !IO->xpr_fread
	|| !IO->xpr_fwrite || !IO->xpr_fseek || !IO->xpr_sread
	|| !IO->xpr_swrite)
    {
	ioerr (IO, "Comm prog missing required function(s); see docs");
	return ERROR;
    }

    if (CheckSetup (IO) == ERROR)
    {
	ioerr (IO, "Setup Fails");
	return ERROR;
    }

    XD->starttime = 0;
    XD->filesize = 0;
    XD->blknum = 0;
    XD->bytes = 0;
    XD->numerrs = 0;
    XD->numtime = 0;
    XD->abort = 0;
    
    if (XD->Kblocks)
    {
	XD->blklen = 1024;
    }
    else
    {
	XD->blklen = 128;
    }
    XD->crcmode = XD->crc_conf;
    return OK;

}

SetTextMode (IO, flag)
PXIO IO;
int flag;
{
    int i;

    switch (XD->Ascii_conf)
    {
	case 'Y':
	    XD->Ascii = 1;
	    break;
	case 'N':
	    XD->Ascii = 0;
	    break;
	case 'C':
	    if (flag && IO->xpr_finfo)
	    {
		switch ((*IO->xpr_finfo) (IO->xpr_filename, 2))
		{
		    case 2:
			XD->Ascii = 1;
			break;
		    case 1:
		    default:
			XD->Ascii = 0;
			break;
		}
	    }
	    else
	    {
		XD->Ascii = 0;
	    }
	    break;
    }
}
char *find_option(buf, option)
char *buf;
char option;
{
    while (*buf)
    {
        buf += strspn(buf," ,\t\r\n");
        if (*buf == option)
        {
            return ++buf;
        }
        buf += strcspn(buf," ,\t\r\n");
    }
    return NULL;
}

ioerr (IO, text)
PXIO IO;
char *text;
{
    XPRU u;
    if(IO->xpr_update)
    {
        setmem(&u, sizeof(XPRU), '\0');
        u.xpru_msg = text;
	u.xpru_updatemask = XPRU_MSG;
	(*IO->xpr_update)(&u);
    }
}
