
#include <stdio.h>
#include <string.h>

#include "xproto.h"
#include "xprymodem.h"

static char *blkchk[] =
{"Checksum", "CRC-16"};
static char *timefmt = "%02ld:%02ld:%02ld";

long XProtocolCleanup (IO)
struct XPR_IO *IO;
{
    if (IO->xpr_data)
    {
	free (IO->xpr_data);
	IO->xpr_data = NULL;
    }
    return (1);
}

DOUpdate (IO, flag)
PXIO IO;
int flag;
{
    XPRU u;
    int secs, elapsed;
    char buff1[20], buff2[20];
    char text[64];

    if (IO->xpr_update)
    {
	setmem (&u, sizeof (XPRU), '\0');

	time (&secs);

	secs -= XD->starttime;

        if(XD->starttime)
	{
	    u.xpru_updatemask |= XPRU_ELAPSEDTIME;

	    sprintf (buff1, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);

	    u.xpru_elapsedtime = buff1;

	    if (secs)
	    {
	        u.xpru_datarate = XD->bytes / secs;
	    }
	    else
	    {
	        u.xpru_datarate = 0;
    	    }

	    if(XD->bytes)
	    {
	    	u.xpru_updatemask |= XPRU_DATARATE;
	    }
	/*
	 * If we know the file size and the data rate we can compute the
	 * estimated time.
	 */

	if (XD->filesize)
	{
	    if (u.xpru_datarate)
	    { 
		u.xpru_updatemask |= XPRU_EXPECTTIME;
		secs = XD->filesize / u.xpru_datarate;
		sprintf (buff2, timefmt, secs / 3600L, (secs / 60L) % 60L, secs % 60L);
		u.xpru_expecttime = buff2;
	    }
	}
        }
	switch (flag)
	{
	    case TIMEOUT:
		u.xpru_blockcheck = XD->crcmode ? blkchk[1] : blkchk[0];
		u.xpru_errormsg = "Timeout";
		u.xpru_timeouts = XD->numtime;
		u.xpru_updatemask |= XPRU_TIMEOUTS | XPRU_ERRORMSG | XPRU_BLOCKCHECK;
		break;
	    case CAN:
		u.xpru_updatemask |= XPRU_ERRORMSG;
		u.xpru_errormsg = "Receiver CANcelled";
		break;
	    case DONE:
		u.xpru_updatemask |= XPRU_MSG;
		u.xpru_msg = "Done";
		break;
	    case OK:
		u.xpru_updatemask |= XPRU_BLOCKS | XPRU_BLOCKSIZE | XPRU_BLOCKCHECK | XPRU_BYTES;
		u.xpru_bytes = XD->bytes;
		u.xpru_blocks = XD->blknum;
		u.xpru_blocksize = XD->blklen;
		u.xpru_blockcheck = XD->crcmode ? blkchk[1] : blkchk[0];
		break;
	    case NAK:
		u.xpru_errors = XD->numerrs;
		u.xpru_errormsg = "Received NAK for sector";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case BURST:
		u.xpru_errors = XD->numerrs;
		u.xpru_errormsg = "Received BURST on ACK";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case WCEOT:
		u.xpru_errors = XD->numerrs;
		u.xpru_errormsg = "No ACK for EOT";
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case CRCERR:
	    case CHKERR:
		u.xpru_errors = XD->numerrs;
		u.xpru_errormsg = text;
		sprintf (text, "%s error at block %d (%d)",
			 (flag == CRCERR) ? "CRC" : "Checksum", XD->blknum, XD->bytes);
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case SEND_START:
		u.xpru_updatemask |= XPRU_BLOCKCHECK | XPRU_MSG;
		u.xpru_msg = text;
		u.xpru_blockcheck = XD->crcmode ? blkchk[1] : blkchk[0];
		sprintf (text, "Sending %s file", (XD->Ascii) ? "Text" : "Binary");
		break;
	    case RECV_START:
		u.xpru_updatemask |= XPRU_BLOCKCHECK | XPRU_MSG;
		u.xpru_msg = text;
		u.xpru_blockcheck = XD->crcmode ? blkchk[1] : blkchk[0];
		sprintf (text, "Receiving %s file", (XD->Ascii) ? "Text" : "Binary");
		break;
	    case ABORTED:
		u.xpru_updatemask |= XPRU_MSG;
		u.xpru_msg = "Transfer Aborted";
		break;
	    case FILERR:
		u.xpru_updatemask |= XPRU_MSG;
		u.xpru_msg = "File open ERROR";
		break;
	    case DUPSEC:
		u.xpru_errormsg = text;
		sprintf (text, "Duplicate Sector %d", XD->blknum);
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case NOSYNC:
		u.xpru_errormsg = "Can't sync ??";
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case STRANGE:
		u.xpru_errormsg = "Totally fucked up";
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    case GARBLE:
		u.xpru_errormsg = text;
		sprintf (text, "Garbled Sector No %d", XD->blknum);
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	    default:
		u.xpru_errormsg = text;
		sprintf (text, "Unexpected character (%d)", flag);
		u.xpru_errors = XD->numerrs;
		u.xpru_updatemask |= XPRU_ERRORS | XPRU_ERRORMSG;
		break;
	}
	(*IO->xpr_update) (&u);
    }
}
