
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "xproto.h"
#include "xprymodem.h"

char _prog_name[] = "<QDOS XPR-YMODEM v2.30>";
#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};

static QVS (_qcfx_,  "QCFX>>01\x00\x09XPRYMODEM\x00\x00\x042.30\xFF\xFF");

char *find_option (char *, char);

static struct xpr_option op[] =
{
    {"YModem options:", XPRO_HEADER, NULL, 0},
    {"Overwrite mode (O,D):", XPRO_STRING, NULL, 0},
    {"CRC mode (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"1K Blocks (Y,N):", XPRO_BOOLEAN, NULL, 0},
    {"Option-G (Y,N):",   XPRO_BOOLEAN, NULL, 0},
    {"Download dir:", XPRO_STRING, NULL, 0}
};

static char option_o[4], option_c[4], option_k[4], option_g[4];

CheckSetup (IO)
PXIO IO;
{
    if (IO->xpr_data == NULL)
    {
	if ((IO->xpr_data = (PXPD) malloc (sizeof (XPD))) == NULL)
	{
	    ioerr (IO, "Not enough memory");
	    return ERROR;
	}
	setmem (IO->xpr_data, sizeof (XPD), '\0');
	XD->crc_conf = 'Y';
	XD->Kblocks_conf = 'Y';
	XD->overwrite = 'O';
	XD->useG = 0;
#ifdef QDOS
	strcpy (XD->DefDir, "ram2_");
#else
	strcpy (XD->DefDir, "t:");
#endif

    }
    return OK;
}

/**********************************************************
 **********************************************************/

long XProtocolSetup (IO)
PXIO IO;
{

    char buf[80], *p;
    int i, len;

    if (CheckSetup (IO) == ERROR)
    {
	return XPRS_FAILURE;
    }

    /* If options string passed by comm prog, use it; else prompt user */
    if (IO->xpr_filename)
    {
	strcpy (buf, IO->xpr_filename);
    }
    else
    {
	/* If xpr_options() implemented by comm program, use it */
	if (IO->xpr_extension >= 1 && IO->xpr_options)
	{
	    op[1].xpro_value = option_o;
	    op[1].xpro_length = sizeof (option_o);
	    *option_o = XD->overwrite;
	    *(option_o + 1) = '\0';
	    op[2].xpro_value = option_c;
	    op[2].xpro_length = sizeof (option_c);
	    *option_c = XD->crc_conf;
	    *(option_c + 1) = '\0';
	    op[3].xpro_value = option_k;
	    op[3].xpro_length = sizeof (option_k);
	    *option_k = XD->Kblocks_conf;
	    *(option_k + 1) = '\0';
	    op[4].xpro_value = option_g;
	    op[4].xpro_length = sizeof (option_g);
	    *option_g = (XD->useG) ? 'Y' : 'N';
	    *(option_g + 1) = '\0';
	    op[5].xpro_value = XD->DefDir;
	    op[5].xpro_length = sizeof (XD->DefDir);

	    /* Convert Y/N used elsewhere into "yes"/"no" required by spec */
	    for (i = 2; i < 5; ++i)
	    {
		strcpy (op[i].xpro_value, (op[i].xpro_value[0] == 'Y') ? "yes" : "no");
	    }

	    (*IO->xpr_options) (op, 6);

	    /* Convert "yes"/"no" or "on"/"off" into Y/N */
	    for (i = 2; i < 5; ++i)
	    {
		if (op[i].xpro_value[0] == 'Y' ||
		    op[i].xpro_value[0] == 'y' ||
		    stricmp (op[i].xpro_value, "yes") == 0 ||
		    stricmp (op[i].xpro_value, "on") == 0)
		{
		    op[i].xpro_value[0] = 'Y';
		}
		else
		{
		    op[i].xpro_value[0] = 'N';
		}

	    }
	    sprintf (buf, "C%c,K%c,O%c,G%c,P%s", *option_c, *option_k,
	                *option_o, *option_g, XD->DefDir);
	}
	else
	{
	    /* Start buffer with current settings so user can see/edit them. */
	    sprintf (buf, "C%c,K%c,O%c,G%c,P%s", XD->crc_conf,
		     XD->Kblocks_conf, XD->overwrite, 
		     (XD->useG) ? 'Y' : 'N', XD->DefDir);
	    if (IO->xpr_gets)
		(*IO->xpr_gets) ("XModem options:", buf);
	}
    }
    /* Upshift options string for easier parsing */
    strupr (buf);

    /*
     * Merge new T(ext) option into current settings if given "TY" = Force
     * Text mode on, "TN" = Force Text mode off, "T?" = Use other end's text
     * mode suggestion (default to binary) "TC" = Ask Comm program for file
     * type
     */
    if (p = find_option (buf, 'O'))
    {
	if (*p == 'O' || *p == 'D')
	    XD->overwrite = *p;
	else
	    ioerr (IO, "Invalid O flag ignored; should be O, D, or S");
    }

    if (p = find_option (buf, 'C'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->crc_conf = *p;
	    XD->crcmode = (XD->crc_conf == 'Y');
	}
	else
	    ioerr (IO, "Invalid C flag ignored; should be Y or N");
    }
    if (p = find_option (buf, 'G'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->useG = (*p == 'Y');
	}
	else
	    ioerr (IO, "Invalid G flag ignored; should be Y or N");
    }

    if (p = find_option (buf, 'K'))
    {
	if (*p == 'Y' || *p == 'N')
	{
	    XD->Kblocks_conf = *p;
	    XD->Kblocks = (XD->Kblocks_conf == 'Y');
	}
	else
	    ioerr (IO, "Invalid K flag ignored; should be Y or N");
    }
    if (p = find_option (buf, 'P'))
    {
	strcpy (XD->DefDir, p);
    }

    return XPRS_NORECREQ | XPRS_SUCCESS;
}

XSetup (IO)
PXIO IO;
{
    if (!IO->xpr_update)
    {
	return ERROR;
    }
    if (!IO->xpr_fopen || !IO->xpr_fclose || !IO->xpr_fread
	|| !IO->xpr_fwrite || !IO->xpr_fseek || !IO->xpr_sread
	|| !IO->xpr_swrite)
    {
	ioerr (IO, "Comm prog missing required function(s); see docs");
	return ERROR;
    }

    if (CheckSetup (IO) == ERROR)
    {
	ioerr (IO, "Setup Fails");
	return ERROR;
    }

    XD->starttime = 0;
    XD->filesize = 0;
    XD->blknum = 0;
    XD->bytes = 0;
    XD->numerrs = 0;
    XD->numtime = 0;
    XD->abort = 0;
    XD->fp = NULL;
    XD->optiong = 0;
    
    if (XD->Kblocks)
    {
	XD->blklen = 1024;
    }
    else
    {
	XD->blklen = 128;
    }
    XD->crcmode = XD->crc_conf;
    return OK;

}

char *find_option (buf, option)
char *buf;
char option;
{
    while (*buf)
    {
	buf += strspn (buf, " ,\t\r\n");
	if (*buf == option)
	{
	    return ++buf;
	}
	buf += strcspn (buf, " ,\t\r\n");
    }
    return NULL;
}

ioerr (IO, text)
PXIO IO;
char *text;
{
    XPRU u;

    if (IO->xpr_update)
    {
	setmem (&u, sizeof (XPRU), '\0');
	u.xpru_msg = text;
	u.xpru_updatemask = XPRU_MSG;
	(*IO->xpr_update) (&u);
    }
}
char *FBuffer (f)
int f;
{
    int n;
    char *v;

    for (n = 16384; n > 1024; n = n >> 2)
    {
	if ((v = malloc ((unsigned) n )))
	{
	    setvbuf ((FILE *) f, v, _IOFBF, (unsigned) n);
	    break;
	}
    }
    return v;
}
